# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test cases for the Controller class.
"""

import unittest

from traits.api import HasTraits, Instance, Str
from traitsui.api import Controller
from traitsui.tests._tools import BaseTestMixin


class FooModel(HasTraits):
    my_str = Str("hallo")


class FooController(Controller):
    """ Test dialog that does nothing useful."""

    model = Instance(FooModel)

    def _model_default(self):
        return FooModel(my_str="meh")


class TestController(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_construction(self):
        # check default constructor.
        dialog = FooController()
        self.assertIsNotNone(dialog.model)
        self.assertEqual(dialog.model.my_str, "meh")

        # check initialization when `model` is explcitly passed in.
        new_model = FooModel()
        dialog = FooController(model=new_model)
        self.assertIs(dialog.model, new_model)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests for the ShadowGroup class.
"""

import unittest

from traitsui.api import Group
from traitsui.group import ShadowGroup

from traitsui.tests._tools import BaseTestMixin


class TestShadowGroup(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_creation_sets_shadow_first(self):
        group = Group()
        # We end up with a DelegationError if the 'shadow' trait is not set
        # first.  Initialization order is dependent on dictionary order, which
        # we can't control, so we throw in a good number of other traits to
        # increase the chance that some other trait is set first.
        ShadowGroup(
            label="dummy",
            show_border=True,
            show_labels=True,
            show_left=True,
            orientation="horizontal",
            scrollable=True,
            shadow=group,
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the Boolean editor factory for all traits toolkit backends.
"""
from traits.api import Dict, Str, Any

from traitsui.editors.text_editor import TextEditor
from traitsui.view import View

# -------------------------------------------------------------------------
#  Trait definitions:
# -------------------------------------------------------------------------

# Mapping from user input text to Boolean values
mapping_trait = Dict(
    Str,
    Any,
    {
        "True": True,
        "true": True,
        "t": True,
        "yes": True,
        "y": True,
        "False": False,
        "false": False,
        "f": False,
        "no": False,
        "n": False,
    },
)


class BooleanEditor(TextEditor):
    """ Editor factory for Boolean editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Dictionary mapping user input to other values.
    #: These definitions override definitions in the 'text_editor' version
    mapping = mapping_trait

    # -------------------------------------------------------------------------
    #  Traits view definition:
    # -------------------------------------------------------------------------

    traits_view = View()

    # -------------------------------------------------------------------------
    #  EditorFactory methods
    # -------------------------------------------------------------------------

    def _get_custom_editor_class(self):
        """ Returns the editor class to use for "custom" style views.
        Overridden to return the simple_editor_class (instead of the
        CustomEditor class for the text editor's factory, which this class
        inherits from).

        """
        return self.simple_editor_class


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = BooleanEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" API for the traitsui package.

Editor Factories
----------------
- :class:`~.BasicEditorFactory`
- :class:`~.EditorFactory`

Context Values
--------------

- :attr:`~.CV`
- :attr:`~.CVFloat`
- :attr:`~.CVInt`
- :attr:`~.CVStr`
- :func:`~.CVType`
- :class:`~.ContextValue`

Editors
-------
- :class:`~.Editor`
- :attr:`~.ArrayEditor`
- :attr:`~.BooleanEditor`
- :attr:`~.ButtonEditor`
- :attr:`~.CheckListEditor`
- :attr:`~.CodeEditor`
- :func:`~.ColorEditor`
- :attr:`~.CompoundEditor`
- :class:`~.CSVListEditor`
- :attr:`~.CustomEditor`
- :class:`~.DateEditor`
- :class:`~.DatetimeEditor`
- :attr:`~.DateRangeEditor`
- :class:`~.DefaultOverride`
- :attr:`~.DirectoryEditor`
- :attr:`~.DNDEditor`
- :attr:`~.DropEditor`
- :attr:`~.EnumEditor`
- :attr:`~.FileEditor`
- :func:`~.FontEditor`
- :attr:`~.HistoryEditor`
- :attr:`~.HTMLEditor`
- :attr:`~.KeyBindingEditor`
- :class:`~.ImageEditor`
- :attr:`~.ImageEnumEditor`
- :attr:`~.InstanceEditor`
- :attr:`~.ListEditor`
- :class:`~.ListStrEditor`
- :attr:`~.NullEditor`
- :class:`~.PopupEditor`
- :attr:`~.ProgressEditor`
- :attr:`~.RangeEditor`
- :func:`~.RGBColorEditor`
- :class:`~.ScrubberEditor`
- :class:`~.SearchEditor`
- :attr:`~.SetEditor`
- :attr:`~.ShellEditor`
- :attr:`~.StyledDateEditor`
- :attr:`~.TableEditor`
- :class:`~.TabularEditor`
- :attr:`~.TextEditor`
- :class:`~.TimeEditor`
- :attr:`~.TitleEditor`
- :attr:`~.TreeEditor`
- :attr:`~.TupleEditor`
- :attr:`~.ValueEditor`

Layout Support
--------------

- :class:`~.Group`
- :class:`~.HFlow`
- :class:`~.HGroup`
- :class:`~.HSplit`
- :class:`~.Tabbed`
- :class:`~.VFlow`
- :class:`~.VFold`
- :class:`~.VGrid`
- :class:`~.VGroup`
- :class:`~.VSplit`

Handlers
--------

- :class:`~.Controller`
- :class:`~.Handler`
- :class:`~.ModelView`
- :class:`~.ViewHandler`
- :func:`~.default_handler`

UI Items
--------

- :class:`~.Custom`
- :class:`~.Heading`
- :class:`~.Item`
- :class:`~.Label`
- :class:`~.Readonly`
- :class:`~.Spring`
- :class:`~.UCustom`
- :class:`~.UItem`
- :class:`~.UReadonly`
- :attr:`~.spring`

Menus and Actions
-----------------

- :class:`~.Action`
- :attr:`~.ActionGroup`
- :attr:`~.ApplyButton`
- :attr:`~.CancelButton`
- :attr:`~.CloseAction`
- :attr:`~.HelpAction`
- :attr:`~.HelpButton`
- :attr:`~.LiveButtons`
- :attr:`~.Menu`
- :attr:`~.MenuBar`
- :attr:`~.ModalButtons`
- :attr:`~.NoButton`
- :attr:`~.NoButtons`
- :attr:`~.OKButton`
- :attr:`~.OKCancelButtons`
- :attr:`~.PyFaceAction`
- :attr:`~.RedoAction`
- :attr:`~.RevertAction`
- :attr:`~.RevertButton`
- :attr:`~.Separator`
- :attr:`~.StandardMenuBar`
- :attr:`~.ToolBar`
- :attr:`~.UndoAction`
- :attr:`~.UndoButton`

Table UI
--------

- :class:`~.TabularAdapter`
- :attr:`~.TableEditor`

Table column types:

- :class:`~.ExpressionColumn`
- :class:`~.ListColumn`
- :class:`~.NumericColumn`
- :class:`~.ObjectColumn`
- :class:`~.TableColumn`

Table filter types:

- :class:`~.EvalTableFilter`
- :class:`~.MenuTableFilter`
- :class:`~.RuleTableFilter`
- :class:`~.TableFilter`

Instance UI
-----------

- :class:`~.InstanceChoice`
- :class:`~.InstanceChoiceItem`
- :class:`~.InstanceDropChoice`
- :attr:`~.InstanceEditor`
- :class:`~.InstanceFactoryChoice`

Toolkit Object
--------------

- :func:`~.toolkit`
- :func:`~.toolkit_object`

Custom Traits
-------------

- :attr:`~.Color`
- :func:`~.ColorTrait`
- :attr:`~.Font`
- :func:`~.FontTrait`
- :attr:`~.RGBColor`
- :func:`~.RGBColorTrait`

Custom UI Traits
----------------

- :attr:`~.Border`
- :attr:`~.HasBorder`
- :attr:`~.HasMargin`
- :attr:`~.Image`
- :attr:`~.Margin`
- :class:`~.StatusItem`

Tree UI
-------

- :class:`~.ITreeNode`
- :class:`~.ITreeNodeAdapter`
- :class:`~.MultiTreeNode`
- :class:`~.ObjectTreeNode`
- :attr:`~.TreeEditor`
- :class:`~.TreeNode`
- :class:`~.TreeNodeObject`

UI and UI Support
-----------------

- :class:`~.UI`
- :class:`~.UIInfo`

Undo Support
------------

- :class:`~.AbstractUndoItem`
- :class:`~.ListUndoItem`
- :class:`~.UndoHistory`
- :class:`~.UndoHistoryUndoItem`
- :class:`~.UndoItem`

View and View Elements
----------------------

- :class:`~.View`
- :class:`~.ViewElement`
- :class:`~.ViewSubElement`
- :mod:`~.view_elements`

Miscellaneous
-------------

- :func:`~.on_help_call`
- :func:`~.help_template`
- :class:`~.Include`
- :func:`~.auto_close_message`
- :func:`~.error`
- :func:`~.message`
- :attr:`~._constants`
- :attr:`~.WindowColor`
- :func:`~.raise_to_debug`

"""

from .basic_editor_factory import BasicEditorFactory

from .context_value import CV, CVFloat, CVInt, CVStr, CVType, ContextValue

from .editor import Editor

from .editor_factory import EditorFactory

try:
    from .editors.api import ArrayEditor
except ImportError:
    # ArrayEditor depends on numpy, so ignore if numpy is not present.
    pass

from .editors.api import (
    BooleanEditor,
    ButtonEditor,
    CheckListEditor,
    CodeEditor,
    ColorEditor,
    CompoundEditor,
    CustomEditor,
    CSVListEditor,
    DNDEditor,
    StyledDateEditor,
    DateEditor,
    DatetimeEditor,
    DateRangeEditor,
    DefaultOverride,
    DirectoryEditor,
    DropEditor,
    EnumEditor,
    FileEditor,
    FontEditor,
    HTMLEditor,
    HistoryEditor,
    ImageEditor,
    ImageEnumEditor,
    InstanceEditor,
    KeyBindingEditor,
    ListEditor,
    ListStrEditor,
    NullEditor,
    PopupEditor,
    ProgressEditor,
    RGBColorEditor,
    RangeEditor,
    ScrubberEditor,
    SearchEditor,
    SetEditor,
    ShellEditor,
    TableEditor,
    TabularEditor,
    TextEditor,
    TimeEditor,
    TitleEditor,
    TreeEditor,
    TupleEditor,
    ValueEditor,
)

from .group import (
    Group,
    HFlow,
    HGroup,
    HSplit,
    Tabbed,
    VFlow,
    VFold,
    VGrid,
    VGroup,
    VSplit,
)

from .handler import (
    Controller,
    Handler,
    ModelView,
    ViewHandler,
    default_handler,
)

from .help import on_help_call

from .help_template import help_template

from .include import Include

from .item import (
    Custom,
    Heading,
    Item,
    Label,
    Readonly,
    Spring,
    UCustom,
    UItem,
    UReadonly,
    spring,
)

from .menu import (
    Action,
    ActionGroup,
    ApplyButton,
    CancelButton,
    CloseAction,
    HelpAction,
    HelpButton,
    LiveButtons,
    Menu,
    MenuBar,
    ModalButtons,
    NoButton,
    NoButtons,
    OKButton,
    OKCancelButtons,
    PyFaceAction,
    RedoAction,
    RevertAction,
    RevertButton,
    Separator,
    StandardMenuBar,
    ToolBar,
    UndoAction,
    UndoButton,
)

from .message import auto_close_message, error, message

from .table_column import (
    ExpressionColumn,
    ListColumn,
    NumericColumn,
    ObjectColumn,
    TableColumn,
)

from .table_filter import (
    EvalTableFilter,
    MenuTableFilter,
    RuleTableFilter,
    TableFilter,
)

from .tabular_adapter import TabularAdapter

from .instance_choice import (
    InstanceChoice,
    InstanceChoiceItem,
    InstanceDropChoice,
    InstanceFactoryChoice,
)

from .toolkit import toolkit, toolkit_object

from .toolkit_traits import (
    Color,
    ColorTrait,
    Font,
    FontTrait,
    RGBColor,
    RGBColorTrait,
)

from .tree_node import (
    ITreeNode,
    ITreeNodeAdapter,
    MultiTreeNode,
    ObjectTreeNode,
    TreeNode,
    TreeNodeObject,
)

from .ui import UI

from .ui_info import UIInfo

from .ui_traits import Border, HasBorder, HasMargin, Image, Margin, StatusItem

from .undo import (
    AbstractUndoItem,
    ListUndoItem,
    UndoHistory,
    UndoHistoryUndoItem,
    UndoItem,
)

from .view import View

from .view_element import ViewElement, ViewSubElement

from . import view_elements

_constants = toolkit().constants()
WindowColor = _constants.get("WindowColor", 0xFFFFFF)


def raise_to_debug():
    """ When we would otherwise silently swallow an exception, call this instead
    to allow people to set the TRAITS_DEBUG environment variable and get the
    exception.
    """
    import os

    if os.getenv("TRAITS_DEBUG") is not None:
        raise



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests for wx.Window's DynamicTargetRegistry
"""

import unittest

from traitsui.testing.api import IsEnabled, IsVisible
from traitsui.testing.tester.exceptions import (
    LocationNotSupported,
)
from traitsui.testing.tester.ui_wrapper import UIWrapper
from traitsui.tests._tools import (
    is_wx,
    requires_toolkit,
    ToolkitName,
)

try:
    import wx
except ImportError:
    if is_wx():
        raise
else:
    from traitsui.testing.tester._ui_tester_registry.wx._control_widget_registry import (   # noqa: E501
        get_widget_registry,
    )


class TargetWithControl:
    """ An object holding a control attribute."""

    def __init__(self, control):
        self.control = control


@requires_toolkit([ToolkitName.wx])
class TestWxControlWidgetRegistry(unittest.TestCase):

    def setUp(self):
        self.widget = wx.Window()
        self.registry = get_widget_registry()
        self.target = TargetWithControl(self.widget)
        self.good_wrapper = UIWrapper(
            target=self.target,
            registries=[self.registry],
        )

    def test_is_enabled(self):
        self.widget.Enable(True)
        self.assertTrue(self.good_wrapper.inspect(IsEnabled()))

    def test_is_visible(self):
        self.widget.Show(True)
        self.assertTrue(self.good_wrapper.inspect(IsVisible()))

    def test_is_invisible(self):
        self.widget.Hide()
        self.assertFalse(self.good_wrapper.inspect(IsVisible()))

    def test_get_interactions_good_target(self):
        self.assertEqual(
            self.registry._get_interactions(self.target),
            set([IsEnabled, IsVisible])
        )

    def test_get_interactions_bad_target(self):
        self.assertEqual(self.registry._get_interactions(None), set())

    def test_get_interaction_doc(self):
        self.assertGreater(
            len(self.registry._get_interaction_doc(self.target, IsEnabled)), 0
        )
        self.assertGreater(
            len(self.registry._get_interaction_doc(self.target, IsVisible)), 0
        )

    def test_get_location_solver(self):
        # There are currently no solvers
        with self.assertRaises(LocationNotSupported):
            self.registry._get_solver(self.target, None)

    def test_get_locations(self):
        self.assertEqual(self.registry._get_locations(self.target), set())

    def test_error_get_location_doc(self):
        with self.assertRaises(LocationNotSupported):
            self.registry._get_location_doc(self.target, None)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""Traits UI 'display only' video editor."""

from traits.api import Bool, Callable, Enum, Float, Instance, Property, Range

from traitsui.context_value import ContextValue, CVType
from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object

AspectRatio = Enum('keep', 'ignore', 'expand')
PlayerState = Enum('stopped', 'playing', 'paused')
MediaStatus = Enum(
    'unknown', 'no_media', 'loading', 'loaded', 'stalled', 'buffering',
    'buffered', 'end', 'invalid',
)


class VideoEditor(BasicEditorFactory):
    """Traits UI 'display only' video editor.

    This editor only displays the video stream, and does not attempt to
    provide UI elements for controlling playback.  It does provide a
    rich set of trait references that can be synchronised with the internal
    state of the video player.
    """

    #: The editor class to be created:
    klass = Property()

    #: The behaviour of the video display when the widget aspect ratio
    #: doesn't match the aspect ratio of the video stream.
    aspect_ratio = AspectRatio()

    #: True if the audio is muted, False otherwise
    muted = CVType(Bool, default=False, sync_value='from')

    #: Audio volume on a logarithmic scale
    volume = CVType(Range(0.0, 100.0), default=75.0, sync_value='from')

    #: The playback speed of the video. Negative values are allowed but may not
    #: be supported by the underlying implementation.
    playback_rate = CVType(Float, default=1.0, sync_value='from')

    #: The state (stopped, playing, paused) of the player
    state = CVType(PlayerState, default='stopped', sync_value='both')

    #: The current position, in seconds, in the video.
    position = CVType(Float, default=0.0, sync_value='both')

    #: Duration of the loaded video in seconds
    duration = Instance(ContextValue, args=('',),
                        allow_none=False, sync_value='to')

    #: The status of the loaded video (see ``MediaStatus``)
    media_status = Instance(ContextValue, args=('',),
                            allow_none=False, sync_value='to')

    #: An integer percentage representing how much of the player's buffer
    #: is filled.
    buffer = Instance(ContextValue, args=('',),
                      allow_none=False, sync_value='to')

    #: A string describing an error encountered by the player
    video_error = Instance(ContextValue, args=('',),
                           allow_none=False, sync_value='to')

    #: Callable to apply to video frames. Takes ref to new frame and a size
    #: tuple. Must return a QImage and a numpy array.
    image_func = CVType(Callable, sync_value='from')

    #: The name of a trait to synchronise with the player's notify interval.
    #: The referenced trait should be a Float representing time in seconds.
    notify_interval = CVType(Float, default=1.0, sync_value='from')

    def _get_klass(self):
        """ Returns the editor class to be instantiated.
        """
        return toolkit_object('video_editor:VideoEditor')



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the compound editor factory for all traits toolkit backends.
"""

from traits.api import Bool, List

from traitsui.editor_factory import EditorFactory

# -------------------------------------------------------------------------
#  Trait definitions:
# -------------------------------------------------------------------------

# List of component editor factories used to build a compound editor
editors_trait = List(EditorFactory)


class CompoundEditor(EditorFactory):
    """ Editor factory for compound editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Component editor factories used to build the editor
    editors = editors_trait

    #: Is user input set on every keystroke?
    auto_set = Bool(True)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = CompoundEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Button editor

A Button trait is displayed as a button in a Traits UI view. When the button is
clicked, Traits UI will execute a method of your choice (a 'listener').

In this example, the listener just increments a click counter.

Please refer to the `ButtonEditor API docs`_ for further information.

.. _ButtonEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.button_editor.html#traitsui.editors.button_editor.ButtonEditor
"""

from traits.api import HasTraits, Button, Int

from traitsui.api import Item, View


class ButtonEditorDemo(HasTraits):
    """ Defines the main ButtonEditor demo class. """

    # Define a Button trait:
    my_button_trait = Button('Click Me')
    click_counter = Int(0)

    # When the button is clicked, do something.
    # The listener method is named '_TraitName_fired', where
    # 'TraitName' is the name of the button trait.
    def _my_button_trait_fired(self):
        self.click_counter += 1

    # Demo view:
    traits_view = View(
        'my_button_trait',
        Item('click_counter', style='readonly'),
        title='ButtonEditor',
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = ButtonEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Edit a string, password, or integer

The TextEditor displays a Str, Password, or Int trait for the user to edit.

The demo shows all styles of the editor for each of the traits, however certain
styles are more useful than others:

- When editing a Str, consider styles 'simple' (one-line), 'custom'
  (multi-line), or 'readonly' (multi-line).
- When editing a Password, style 'simple' is recommended (shows asterisks).
- When editing an Int, consider styles 'simple' and 'readonly'.

Please refer to the `TextEditor API docs`_ for further information.

.. _TextEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.text_editor.html#traitsui.editors.text_editor.TextEditor
"""

from traits.api import HasTraits, Str, Int, Password

from traitsui.api import Item, Group, View


# The main demo class:
class TextEditorDemo(HasTraits):
    """ Defines the TextEditor demo class.
    """

    # Define a trait for each of three TextEditor variants:
    string_trait = Str("sample string")
    int_trait = Int(1)
    password = Password()

    # TextEditor display with multi-line capability (for a string):
    text_str_group = Group(
        Item('string_trait', style='simple', label='Simple'),
        Item('_'),
        Item('string_trait', style='custom', label='Custom'),
        Item('_'),
        Item('string_trait', style='text', label='Text'),
        Item('_'),
        Item('string_trait', style='readonly', label='ReadOnly'),
        label='String'
    )

    # TextEditor display without multi-line capability (for an integer):
    text_int_group = Group(
        Item('int_trait', style='simple', label='Simple', id="simple_int"),
        Item('_'),
        Item('int_trait', style='custom', label='Custom', id="custom_int"),
        Item('_'),
        Item('int_trait', style='text', label='Text', id="text_int"),
        Item('_'),
        Item(
            'int_trait', style='readonly', label='ReadOnly', id="readonly_int",
        ),
        label='Integer'
    )

    # TextEditor display with secret typing capability (for Password traits):
    text_pass_group = Group(
        Item('password', style='simple', label='Simple'),
        Item('_'),
        Item('password', style='custom', label='Custom'),
        Item('_'),
        Item('password', style='text', label='Text'),
        Item('_'),
        Item('password', style='readonly', label='ReadOnly'),
        label='Password'
    )

    # The view includes one group per data type. These will be displayed
    # on separate tabbed panels:
    traits_view = View(
        text_str_group,
        text_pass_group,
        text_int_group,
        title='TextEditor',
        buttons=['OK']
    )


# Create the demo:
demo = TextEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == "__main__":
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Implementation of a TableEditor demo plugin for Traits UI demo program

This demo shows the full behavior of a straightforward TableEditor.  Only
one style of TableEditor is implemented, so that is the one shown.

Please refer to the `TableEditor API docs`_ for further information.

.. _TableEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.table_editor.html#traitsui.editors.table_editor.TableEditor
"""
# Issue related to the demo warning: enthought/traitsui#948


from traits.api import HasTraits, HasStrictTraits, Str, Int, Regex, List

from traitsui.api import (
    View, Group, Item, TableEditor, ObjectColumn, ExpressionColumn,
    EvalTableFilter
)
from traitsui.table_filter import (
    EvalFilterTemplate, MenuFilterTemplate, RuleFilterTemplate
)


# A helper class for the 'Department' class below:
class Employee(HasTraits):
    first_name = Str()
    last_name = Str()
    age = Int()
    phone = Regex(value='000-0000', regex=r'\d\d\d[-]\d\d\d\d')

    traits_view = View(
        'first_name', 'last_name', 'age', 'phone',
        title='Create new employee',
        width=0.18,
        buttons=['OK', 'Cancel']
    )


# For readability, the TableEditor of the demo is defined here, rather than in
# the View:
table_editor = TableEditor(
    columns=[
        ObjectColumn(name='first_name', width=0.20),
        ObjectColumn(name='last_name', width=0.20),
        ExpressionColumn(
            label='Full Name',
            width=0.30,
            expression="'%s %s' % (object.first_name, "
            "object.last_name )"
        ),
        ObjectColumn(name='age', width=0.10, horizontal_alignment='center'),
        ObjectColumn(name='phone', width=0.20)
    ],
    deletable=True,
    sort_model=True,
    auto_size=False,
    orientation='vertical',
    edit_view=View(
        Group('first_name', 'last_name', 'age', 'phone', show_border=True),
        resizable=True
    ),
    filters=[EvalFilterTemplate, MenuFilterTemplate, RuleFilterTemplate],
    search=EvalTableFilter(),
    show_toolbar=True,
    row_factory=Employee,
)


# The class to be edited with the TableEditor:
class Department(HasStrictTraits):

    employees = List(Employee)

    traits_view = View(
        Group(
            Item(
                'employees',
                show_label=False,
                editor=table_editor
            ),
            show_border=True,
        ),
        title='Department Personnel',
        width=.4,
        height=.4,
        resizable=True,
        buttons=['OK'],
        kind='live'
    )


# Create some employees:
employees = [
    Employee(first_name='Jason', last_name='Smith',
             age=32, phone='555-1111'),
    Employee(first_name='Mike', last_name='Tollan',
             age=34, phone='555-2222'),
    Employee(first_name='Dave', last_name='Richards',
             age=42, phone='555-3333'),
    Employee(first_name='Lyn', last_name='Spitz',
             age=40, phone='555-4444'),
    Employee(first_name='Greg', last_name='Andrews',
             age=45, phone='555-5555')
]

# Create the demo:
demo = Department(employees=employees)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module contains targets for UIWrapper so that the logic related to
them can be reused.
To use the logic in these objects, if the class in this module is a base class
(indicated by leading _Base) they simply need to subclass this class, override
any necessary traits, and then call the register method.
"""


class BaseSourceWithLocation:
    """ Wrapper base class to hold locator information together with a source
    (typically an editor).  This is useful for cases in which the location
    information is still necessary when performing actions such as a mouse
    click or key click.

    For example, an Enum editor and an index.
    This class is meant to be subclassed for specific usecases, and the
    class level attributes overridden.
    """

    # The type of the source object on which the location information will be
    # evaluated on
    source_class = None
    # The type of the locator object that provides location information.
    # (e.g. locator.Index)
    locator_class = None
    # the handlers we want to register for the given source_class
    # must be given as a list of tuples where the first element is the
    # interaction class (e.g. command.MouseClick) and the second is the
    # actual handler function.
    # See TargetRegistry.register_interaction
    # for the signature of the callable.
    handlers = []

    def __init__(self, source, location):
        """
        Parameters
        ----------
        source : instance of source_class
            The source object. Typically this is an editor.
        location : instance of locator_class
            The location information of interest
        """
        self.source = source
        self.location = location

    @classmethod
    def register(cls, registry):
        """ Class method to register interactions on a
        _SourceWithLocation for the given registry. It is expected that this
        class method will be called by subclasses, and thus interactions would
        be registered to subclasses rather than the base class.

        If there are any conflicts, an error will occur.

        Parameters
        ----------
        registry : TargetRegistry
            The registry being registered to.
        """
        registry.register_location(
            target_class=cls.source_class,
            locator_class=cls.locator_class,
            solver=lambda wrapper, location: cls(wrapper._target, location),
        )
        for interaction_class, handler in cls.handlers:
            registry.register_interaction(
                target_class=cls,
                interaction_class=interaction_class,
                handler=handler
            )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module contains functions used by toolkit specific implementation for
normalizing differences among toolkits (Qt and Wx).
"""


def check_key_compat(key):
    """ Check if the given key is a unicode character within the range of
    values currently supported for emulating key sequences on both Qt and Wx
    textboxes.

    Parameters
    ----------
    key : str
        A unicode character

    Raises
    ------
    ValueError
        If the unicode character is not within the supported range of values.
    """
    # Support for more characters can be added when there are needs.
    if ord(key) < 32 or ord(key) >= 127:
        raise ValueError(
            f"Key {key!r} is currently not supported. "
            f"Supported characters between code point 32 - 126."
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Editor factory for scrubber-based integer or float value editors.
"""

from pyface.ui_traits import Alignment
from traits.api import Float, Property

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object
from traitsui.toolkit_traits import Color

# -------------------------------------------------------------------------
#  Create the editor factory object:
# -------------------------------------------------------------------------

# Editor factory for scrubber-based integer or float value editors.


class ScrubberEditor(BasicEditorFactory):

    #: The editor class to be instantiated:
    klass = Property()

    #: The low end of the scrubber range:
    low = Float()

    #: The high end of the scrubber range:
    high = Float()

    #: The normal increment (default: auto-calculate):
    increment = Float()

    #: The alignment of the text within the scrubber:
    alignment = Alignment("center")

    #: The background color for the scrubber:
    color = Color(None)

    #: The hover mode background color for the scrubber:
    hover_color = Color(None)

    #: The active mode background color for the scrubber:
    active_color = Color(None)

    #: The scrubber border color:
    border_color = Color(None)

    #: The color to use for the value text:
    text_color = Color("black")

    def _get_klass(self):
        """ Returns the toolkit-specific editor class to be instantiated.
        """
        return toolkit_object("scrubber_editor:_ScrubberEditor")



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various button editors for the wxPython user interface toolkit.
"""


import wx

from pyface.ui_traits import Image
from traits.api import Str, observe

from .editor import Editor

# -------------------------------------------------------------------------
#  'SimpleEditor' class:
# -------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simple style editor for a button.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The button label
    label = Str()

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        label = self.factory.label or self.item.get_label(self.ui)
        self.control = wx.Button(parent, -1, self.string_value(label))
        self.sync_value(self.factory.label_value, "label", "from")
        self.control.Bind(wx.EVT_BUTTON, self.update_object)
        self.set_tooltip()

    def _label_changed(self, label):
        self.control.SetLabel(self.string_value(label))

    def update_object(self, event):
        """ Handles the user clicking the button by setting the factory value
            on the object.
        """
        factory = self.factory
        self.value = factory.value

        # If there is an associated view, then display it:
        if factory.view is not None:
            self.object.edit_traits(view=factory.view, parent=self.control)

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        self.control.Unbind(wx.EVT_BUTTON)
        super().dispose()


class CustomEditor(SimpleEditor):
    """ Custom style editor for a button, which can contain an image.
    """
    #: The button image
    image = Image()

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        from pyface.ui.wx.image_button import ImageButton

        factory = self.factory
        if self.factory.label:
            label = self.factory.label
        else:
            label = self.item.get_label(self.ui)

        self._control = ImageButton(
            parent,
            label=self.string_value(label),
            image=factory.image,
            style=factory.style,
            orientation=factory.orientation,
            width_padding=factory.width_padding,
            height_padding=factory.height_padding,
        )
        self.control = self._control.control
        self._control.on_trait_change(
            self.update_object, "clicked", dispatch="ui"
        )
        self.sync_value(self.factory.label_value, "label", "from")
        self.sync_value(self.factory.image_value, "image", "from")
        self.set_tooltip()

    def _label_changed(self, label):
        self._control.label = self.string_value(label)

    @observe("image")
    def _image_updated(self, event):
        image = event.new
        self._control.image = image

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        self._control.on_trait_change(
            self.update_object, "clicked", remove=True
        )

        super().dispose()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


try:
    from traitsui._version import full_version as __version__
except ImportError:
    __version__ = "not-built"

__requires__ = ["traits>=6.2.0", "pyface>=7.3.0"]
__extras_require__ = {
    "wx": ["wxpython>=4", "numpy"],
    "pyqt": ["pyqt>=4.10", "pygments"],
    "pyqt5": ["pyqt5", "pygments"],
    "pyside2": ["pyside2", "shiboken2", "pygments"],
    "demo": [
        # to be deprecated, see enthought/traitsui#950
        "configobj", "docutils",
    ],
    "examples": [
        # Dependencies for examples
        "apptools",
        "chaco",   # for a very simple example, see enthought/traitsui#1139
        "h5py",
        "numpy",
        "pandas",
        "tables",
    ],
    "editors": [
        # Optional dependencies for certain editors which may not be needed by
        # projects. If they are absent, ``traitsui.api``` should still be
        # importable and the relevant tests should be skipped.
        "numpy",   # For ArrayEditor and DataFrameEditor
        "pandas",  # For DataFrameEditor
    ],
    "test": [
        # Dependencies for running test suites.
        "packaging",
        "numpy"
    ],
}


# ============================= Test Loader ==================================
def load_tests(loader, standard_tests, pattern):
    """ Custom test loading function that enables test filtering using regex
    exclusion pattern.

    Parameters
    ----------
    loader : unittest.TestLoader
        The instance of test loader
    standard_tests : unittest.TestSuite
        Tests that would be loaded by default from this module (no tests)
    pattern : str
        An inclusion pattern used to match test files (test*.py default)

    Returns
    -------
    filtered_package_tests : unittest.TestSuite
        TestSuite representing all package tests that did not match specified
        exclusion pattern.
    """
    from os.path import dirname
    from unittest import TestSuite
    from traits.etsconfig.api import ETSConfig
    from traitsui.tests._tools import filter_tests

    # Make sure the right toolkit is up and running before importing tests
    from traitsui.toolkit import toolkit
    toolkit()

    if ETSConfig.toolkit.startswith("qt"):
        exclusion_pattern = "wx"
    elif ETSConfig.toolkit == "wx":
        exclusion_pattern = "qt"
    else:
        exclusion_pattern = "(wx|qt)"

    this_dir = dirname(__file__)
    package_tests = loader.discover(start_dir=this_dir, pattern=pattern)

    if exclusion_pattern is None:
        return package_tests

    filtered_package_tests = TestSuite()
    for test_suite in package_tests:
        filtered_test_suite = filter_tests(test_suite, exclusion_pattern)
        filtered_package_tests.addTest(filtered_test_suite)

    return filtered_package_tests



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.ui_editors.array_view_editor import (
    _ArrayViewEditor as BaseArrayViewEditor,
)

from .ui_editor import UIEditor

# -------------------------------------------------------------------------
#  '_ArrayViewEditor' class:
# -------------------------------------------------------------------------


class _ArrayViewEditor(BaseArrayViewEditor, UIEditor):
    pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module defines interaction objects that can be passed to
``UIWrapper.inspect`` where the actions represent 'queries'.

Implementations for these actions are expected to return value(s), ideally
without incurring side-effects.
"""


class SelectedText:
    """ An object representing an interaction to obtain the displayed (echoed)
    plain text which is currently selected.

    E.g. For a Enum List, with one entry currently selected, the displayed
    selected text would be the label of that entry.

    Implementations should return a ``str``, or None if nothing is selected.
    """
    pass


class DisplayedText:
    """ An object representing an interaction to obtain the displayed (echoed)
    plain text.

    E.g. For a textbox using a password styling, the displayed text should
    be a string of platform-dependent password mask characters.

    Implementations should return a ``str``.
    """
    pass


class IsChecked:
    """ An object representing an interaction to obtain whether a checkable
    widget (e.g. checkbox) is checked or not.

    Implementations should return True if checked and False if not.
    """
    pass


class IsEnabled:
    """ An object representing an interaction to obtain whether a widget is
    enabled or not.

    Implementations should return True if enabled and False if not.
    """
    pass


class IsVisible:
    """ An object representing an interaction to obtain whether a widget is
    visible or not.

    Implementations should return True if visible and False if not.
    """
    pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Traits UI 'display only' image editor.
"""

from pyface.ui_traits import Image
from traits.api import Bool, Property

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object

# -------------------------------------------------------------------------
#  'ImageEditor' editor factory class:
# -------------------------------------------------------------------------


class ImageEditor(BasicEditorFactory):

    #: The editor class to be created:
    klass = Property()

    #: The optional image resource to be displayed by the editor (if not
    #: specified, the editor's object value is used as the ImageResource to
    #: display):
    image = Image

    #: The following traits are currently supported on Qt only

    #: Whether or not to scale the image to fit the available space
    scale = Bool()

    #: Whether or not to scale the image larger than the original when scaling
    allow_upscaling = Bool()

    #: Whether or not to preserve the aspect ratio when scaling
    preserve_aspect_ratio = Bool()

    #: Whether or not to allow the image to be clipped when not scaling
    allow_clipping = Bool()

    def _get_klass(self):
        """ Returns the editor class to be instantiated.
        """
        return toolkit_object("image_editor:_ImageEditor")



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

from traitsui.testing.tester import command
from traitsui.testing.tester.exceptions import Disabled
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers

from traitsui.tests._tools import (
    is_wx,
    requires_toolkit,
    ToolkitName,
)

try:
    import wx
except ImportError:
    if is_wx():
        raise


@requires_toolkit([ToolkitName.wx])
class TestInteractions(unittest.TestCase):

    def setUp(self):
        self.frame = wx.Frame(None)
        self.frame.Show()

    def tearDown(self):
        self.frame.Close()
        self.frame.Destroy()

    def test_mouse_click(self):
        handler = mock.Mock()
        button = wx.Button(self.frame)
        button.Bind(wx.EVT_BUTTON, handler)

        # when
        _interaction_helpers.mouse_click_button(control=button, delay=0)

        # then
        self.assertEqual(handler.call_count, 1)

    def test_mouse_click_disabled_button(self):
        handler = mock.Mock()
        button = wx.Button(self.frame)
        button.Bind(wx.EVT_BUTTON, handler)
        button.Enable(False)

        # when
        _interaction_helpers.mouse_click_button(control=button, delay=0)

        # then
        self.assertEqual(handler.call_count, 0)

    def test_mouse_click_None_warns(self):
        control = None
        with self.assertWarns(UserWarning):
            _interaction_helpers.mouse_click_button(control=control, delay=0)

    def test_key_sequence(self):
        # The insertion point is moved to the end
        textbox = wx.TextCtrl(self.frame)
        textbox.SetValue("123")
        handler = mock.Mock()
        textbox.Bind(wx.EVT_TEXT, handler)

        _interaction_helpers.key_sequence_text_ctrl(
            textbox, command.KeySequence("abc"), 0)

        self.assertEqual(textbox.GetValue(), "123abc")
        self.assertEqual(handler.call_count, 3)

    def test_key_sequence_with_unicode(self):
        handler = mock.Mock()
        textbox = wx.TextCtrl(self.frame)
        textbox.Bind(wx.EVT_TEXT, handler)
        # This range is supported by Qt
        for code in range(32, 127):
            with self.subTest(code=code, word=chr(code)):
                textbox.Clear()
                handler.reset_mock()

                # when
                _interaction_helpers.key_sequence_text_ctrl(
                    textbox,
                    command.KeySequence(chr(code) * 3),
                    delay=0,
                )

                # then
                self.assertEqual(textbox.Value, chr(code) * 3)
                self.assertEqual(handler.call_count, 3)

    def test_key_sequence_with_backspace_unsupported(self):
        textbox = wx.TextCtrl(self.frame)

        with self.assertRaises(ValueError) as exception_context:
            _interaction_helpers.key_sequence_text_ctrl(
                textbox, command.KeySequence("\b"), 0
            )

        self.assertIn(
            "is currently not supported.",
            str(exception_context.exception),
        )

    def test_key_sequence_disabled(self):
        textbox = wx.TextCtrl(self.frame)
        textbox.SetEditable(False)

        with self.assertRaises(Disabled):
            _interaction_helpers.key_sequence_text_ctrl(
                textbox, command.KeySequence("abc"), 0)

    def test_key_click(self):
        textbox = wx.TextCtrl(self.frame)
        handler = mock.Mock()
        textbox.Bind(wx.EVT_TEXT, handler)

        _interaction_helpers.key_click_text_entry(
            textbox, command.KeyClick("A"), 0)

        self.assertEqual(textbox.Value, "A")
        self.assertEqual(handler.call_count, 1)

    def test_key_click_backspace(self):
        textbox = wx.TextCtrl(self.frame)
        textbox.SetValue("A")
        handler = mock.Mock()
        textbox.Bind(wx.EVT_TEXT, handler)

        _interaction_helpers.key_click_text_entry(
            textbox, command.KeyClick("Backspace"), 0)

        self.assertEqual(textbox.Value, "")
        self.assertEqual(handler.call_count, 1)

    def test_key_click_backspace_with_selection(self):
        textbox = wx.TextCtrl(self.frame)
        textbox.SetFocus()
        textbox.SetValue("ABCDE")
        textbox.SetSelection(0, 4)
        # sanity check
        self.assertEqual(textbox.GetStringSelection(), "ABCD")

        handler = mock.Mock()
        textbox.Bind(wx.EVT_TEXT, handler)

        _interaction_helpers.key_click_text_entry(
            textbox, command.KeyClick("Backspace"), 0)

        self.assertEqual(textbox.Value, "E")
        self.assertEqual(handler.call_count, 1)

    def test_key_click_end(self):
        textbox = wx.TextCtrl(self.frame)
        textbox.SetValue("ABCDE")
        textbox.SetInsertionPoint(0)

        # sanity check
        self.assertEqual(textbox.GetInsertionPoint(), 0)

        _interaction_helpers.key_click_text_entry(
            textbox, command.KeyClick("End"), 0)
        _interaction_helpers.key_click_text_entry(
            textbox, command.KeyClick("F"), 0)

        self.assertEqual(textbox.Value, "ABCDEF")

    def test_key_click_disabled(self):
        textbox = wx.TextCtrl(self.frame)
        textbox.SetEditable(False)

        with self.assertRaises(Disabled):
            _interaction_helpers.key_click_text_entry(
                textbox, command.KeyClick("Enter"), 0)

    def test_key_click_slider_helpful_err(self):
        slider = wx.Slider()
        with self.assertRaises(ValueError) as exc:
            _interaction_helpers.key_click_slider(
                slider, command.KeyClick("Enter"), 0)
        self.assertIn(
            "['Down', 'Left', 'Page Down', 'Page Up', 'Right', 'Up']",
            str(exc.exception)
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Auto-update from a list to a TabularEditor

Demonstrates using a TabularEditor with the 'auto_update' feature enabled,
which allows the tabular editor to automatically update itself when the content
of any object in the list associated with the editor is modified.

To interact with the demo:

- Select an employee from the list.
- Adjust their salary increase.
- Click the **Give raise** button.
- Observe that the table automatically updates to reflect the employees new
  salary.

In order for auto-update to work correctly, the editor trait should be a list
of objects derived from HasTraits. Also, performance can be affected when very
long lists are used, since enabling this feature adds and removed Traits
listeners to each item in the list.

"""
# Issues related to the demo warning:
# enthought/traitsui#960

from traits.api import HasTraits, Str, Float, List, Instance, Button
from traitsui.api import (
    View, HGroup, Item, TabularAdapter, TabularEditor, spring
)


# -- EmployeeAdapter Class ------------------------------------------------
class EmployeeAdapter(TabularAdapter):

    columns = [('Name', 'name'), ('Salary', 'salary')]

    def get_default_value(self, object, trait):
        return Employee(name="John", salary=30000)


# -- Employee Class -------------------------------------------------------
class Employee(HasTraits):

    name = Str()
    salary = Float()


# -- Company Class --------------------------------------------------------
class Company(HasTraits):

    employees = List(Employee)
    employee = Instance(Employee)
    increase = Float()
    give_raise = Button('Give raise')

    traits_view = View(
        Item(
            'employees',
            show_label=False,
            editor=TabularEditor(
                adapter=EmployeeAdapter(),
                selected='employee',
                auto_resize=True,
                auto_update=True
            )
        ),
        HGroup(
            spring,
            Item('increase'),
            Item(
                'give_raise',
                show_label=False,
                enabled_when='employee is not None'
            )
        ),
        title='Auto Update Tabular Editor demo',
        height=0.25,
        width=0.30,
        resizable=True
    )

    def _give_raise_changed(self):
        self.employee.salary += self.increase
        self.employee = None


# -- Set up the demo ------------------------------------------------------
demo = Company(
    increase=1000,
    employees=[
        Employee(name='Fred', salary=45000),
        Employee(name='Sally', salary=52000),
        Employee(name='Jim', salary=39000),
        Employee(name='Helen', salary=41000),
        Employee(name='George', salary=49000),
        Employee(name='Betty', salary=46000)
    ]
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Implementation of a DirectoryEditor demo plugin for Traits UI demo program.

This demo shows each of the four styles of the DirectoryEditor.

Please refer to the `DirectoryEditor API docs`_ for further information.

.. _DirectoryEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.directory_editor.html#traitsui.editors.directory_editor.DirectoryEditor
"""
# Issue related to the demo warning: enthought/traitsui#889


from traits.api import HasTraits, Directory

from traitsui.api import Item, Group, View


# Define the demo class:
class DirectoryEditorDemo(HasTraits):
    """ Define the main DirectoryEditor demo class. """

    # Define a Directory trait to view:
    dir_name = Directory()

    # Display specification (one Item per editor style):
    dir_group = Group(
        Item('dir_name', style='simple', label='Simple'),
        Item('_'),
        Item('dir_name', style='custom', label='Custom'),
        Item('_'),
        Item('dir_name', style='text', label='Text'),
        Item('_'),
        Item('dir_name', style='readonly', label='ReadOnly')
    )

    # Demo view:
    traits_view = View(
        dir_group,
        title='DirectoryEditor',
        width=400,
        height=600,
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = DirectoryEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This is a modified version of the Standard_Editors/TreeEditor_demo.py in which
the `Employee` class is now a subclass of HasRequiredTraits.  Thus, to create a
new `TreeNode` for `Employee`, to create the associated `Employee` object we
can not simply do `Employee()`. Luckily the `TreeNode` API provides a means for
specifying a callable that returns a created instance when we want to add a new
node. To do so we can simply pass a tuple of the for (klass, prompt, factorry)
as an item the `add` list trait for a TreeNode. klass is the class of object we
want to be addable, prompt is a boolean indicating whether or not we want to
prompt the user to specify traits after we instantiate the object before it is
added to the tree, and factory is the previously described callable that must
return an instance of klass.
"""

from traits.api import HasRequiredTraits, HasTraits, Str, Regex, List, Instance

from traitsui.api import Item, View, TreeEditor, TreeNode


class Employee(HasRequiredTraits):
    """ Defines a company employee. """

    name = Str('<unknown>')
    title = Str()
    phone = Regex(regex=r'\d\d\d-\d\d\d\d')

    def default_title(self):
        self.title = 'Senior Engineer'


def create_default_employee():
    return Employee(name="Dilbert", title="Engineer", phone="999-8212")


class Department(HasTraits):
    """ Defines a department with employees. """

    name = Str('<unknown>')
    employees = List(Employee)


class Company(HasTraits):
    """ Defines a company with departments and employees. """

    name = Str('<unknown>')
    departments = List(Department)
    employees = List(Employee)


# Create an empty view for objects that have no data to display:
no_view = View()

# Define the TreeEditor used to display the hierarchy:
tree_editor = TreeEditor(
    nodes=[
        # The first node specified is the top level one
        TreeNode(
            node_for=[Company],
            auto_open=True,
            # child nodes are
            children='',
            label='name',  # label with Company name
            view=View(['name'])
        ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='departments',
            label='=Departments',  # constant label
            view=no_view,
            add=[Department],
         ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='employees',
            label='=Employees',   # constant label
            view=no_view,
            add=[(Employee, True, create_default_employee)]
        ),
        TreeNode(
            node_for=[Department],
            auto_open=True,
            children='employees',
            label='name',   # label with Department name
            view=View(['name']),
            add=[(Employee, True, create_default_employee)]
        ),
        TreeNode(
            node_for=[Employee],
            auto_open=True,
            label='name',   # label with Employee name
            view=View(['name', 'title', 'phone'])
        )
    ]
)


class Partner(HasTraits):
    """ Defines a business partner."""

    name = Str('<unknown>')
    company = Instance(Company)

    traits_view = View(
        Item(name='company', editor=tree_editor, show_label=False),
        title='Company Structure',
        buttons=['OK'],
        resizable=True,
        style='custom',
        width=.3,
        height=500
    )


# Create an example data structure:
jason = Employee(name='Jason', title='Senior Engineer', phone='536-1057')
mike = Employee(name='Mike', title='Senior Engineer', phone='536-1057')
dave = Employee(name='Dave', title='Senior Developer', phone='536-1057')
martin = Employee(name='Martin', title='Senior Engineer', phone='536-1057')
duncan = Employee(name='Duncan', title='Consultant', phone='526-1057')

# Create the demo:
demo = Partner(
    name='Enthought, Inc.',
    company=Company(
        name='Enthought',
        employees=[dave, martin, duncan, jason, mike],
        departments=[
            Department(
                name='Business',
                employees=[jason, mike]
            ),
            Department(
                name='Scientific',
                employees=[dave, martin, duncan]
            )
        ]
    )
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various instance descriptors used by the instance editor and
    instance editor factory classes.
"""

from traits.api import HasPrivateTraits, Str, Any, Dict, Tuple, Callable, Bool

from .ui_traits import AView

from .helper import user_name_for


class InstanceChoiceItem(HasPrivateTraits):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: User interface name for the item
    name = Str()

    #: View associated with this item
    view = AView

    #: Does this item create new instances?
    is_factory = Bool(False)

    def get_name(self, object=None):
        """ Returns the name of the item.
        """
        return self.name

    def get_view(self):
        """ Returns the view associated with the object.
        """
        return self.view

    def get_object(self):
        """ Returns the object associated with the item.
        """
        raise NotImplementedError

    def is_compatible(self, object):
        """ Indicates whether a specified object is compatible with the item.
        """
        raise NotImplementedError

    def is_selectable(self):
        """ Indicates whether the item can be selected by the user.
        """
        return True

    def is_droppable(self):
        """ Indicates whether the item supports drag and drop.
        """
        return False


class InstanceChoice(InstanceChoiceItem):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Object associated with the item
    object = Any()

    #: The name of the object trait containing its user interface name:
    name_trait = Str("name")

    def get_name(self, object=None):
        """ Returns the name of the item.
        """
        if self.name != "":
            return self.name

        name = getattr(self.object, self.name_trait, None)
        if isinstance(name, str):
            return name

        return user_name_for(self.object.__class__.__name__)

    def get_object(self):
        """ Returns the object associated with the item.
        """
        return self.object

    def is_compatible(self, object):
        """ Indicates whether a specified object is compatible with the item.
        """
        return object is self.object


class InstanceFactoryChoice(InstanceChoiceItem):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Indicates whether an instance compatible with this item can be dragged
    #: and dropped rather than created
    droppable = Bool(False)

    #: Indicates whether the item can be selected by the user
    selectable = Bool(True)

    #: A class (or other callable) that can be used to create an item
    #: compatible with this item
    klass = Callable()

    #: Tuple of arguments to pass to **klass** to create an instance
    args = Tuple()

    #: Dictionary of arguments to pass to **klass** to create an instance
    kw_args = Dict(Str, Any)

    #: Does this item create new instances? This value overrides the default.
    is_factory = True

    def get_name(self, object=None):
        """ Returns the name of the item.
        """
        if self.name != "":
            return self.name

        name = getattr(object, "name", None)
        if isinstance(name, str):
            return name

        if issubclass(type(self.klass), type):
            klass = self.klass
        else:
            klass = self.get_object().__class__

        return user_name_for(klass.__name__)

    def get_object(self):
        """ Returns the object associated with the item.
        """
        return self.klass(*self.args, **self.kw_args)

    def is_droppable(self):
        """ Indicates whether the item supports drag and drop.
        """
        return self.droppable

    def is_compatible(self, object):
        """ Indicates whether a specified object is compatible with the item.
        """
        if issubclass(type(self.klass), type):
            return isinstance(object, self.klass)
        return isinstance(object, self.get_object().__class__)

    def is_selectable(self):
        """ Indicates whether the item can be selected by the user.
        """
        return self.selectable


class InstanceDropChoice(InstanceFactoryChoice):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Indicates whether an instance compatible with this item can be dragged
    #: and dropped rather than created . This value overrides the default.
    droppable = True

    #: Indicates whether the item can be selected by the user. This value
    #: overrides the default.
    selectable = False

    #: Does this item create new instances? This value overrides the default.
    is_factory = False



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test cases for the TreeNode object.
"""

import unittest

from traits.api import HasStrictTraits, List, Str, This
from traits.testing.api import UnittestTools

from traitsui.api import TreeNode
from traitsui.tests._tools import BaseTestMixin


class DummyModel(HasStrictTraits):
    """ Dummy model with children.
    """

    name = Str()

    children = List(This)


class TestTreeNode(BaseTestMixin, UnittestTools, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_insert_child(self):
        # Regression test for #559
        model = DummyModel(
            name="Parent",
            children=[DummyModel(name="Child0"), DummyModel(name="Child2")],
        )
        node = TreeNode(children="children", node_for=[DummyModel])
        node.insert_child(model, 1, DummyModel(name="Child1"))

        # Assert
        self.assertEqual(len(model.children), 3)
        for i in range(3):
            self.assertEqual(model.children[i].name, "Child{}".format(i))


if __name__ == "__main__":
    unittest.run()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Tree editor for hierarchal data

Demonstrates using the TreeEditor to display a hierarchically organized data
structure.

In this case, the tree has the following hierarchy:

- Partner
    - Company
        - Department
            - Employee

The TreeEditor generates a hierarchical tree control, consisting of nodes. It
is useful for cases where objects contain lists of other objects.

The tree control is displayed in one pane of the editor, and a user interface
for the selected object is displayed in the other pane. The layout orientation
of the tree and the object editor is determined by the *orientation* parameter
of TreeEditor(), which can be 'horizontal' or 'vertical'.

You must specify the types of nodes that can appear in the tree using the
*nodes* parameter, which must be a list of instances of TreeNode (or of
subclasses of TreeNode).

You must specify the classes whose instances the node type applies to. Use the
**node_for** attribute of TreeNode to specify a list of classes; often, this
list contains only one class. You can have more than one node type that applies
to a particular class; in this case, each object of that class is represented
by multiple nodes, one for each applicable node type.

See the Traits User Manual for more details.

Please refer to the `TreeEditor API docs`_ for further information.

.. _TreeEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.tree_editor.html#traitsui.editors.tree_editor.TreeEditor
"""

from traits.api import HasTraits, Str, Regex, List, Instance

from traitsui.api import Item, View, TreeEditor, TreeNode


class Employee(HasTraits):
    """ Defines a company employee. """

    name = Str('<unknown>')
    title = Str()
    phone = Regex(regex=r'\d\d\d-\d\d\d\d')

    def default_title(self):
        self.title = 'Senior Engineer'


class Department(HasTraits):
    """ Defines a department with employees. """

    name = Str('<unknown>')
    employees = List(Employee)


class Company(HasTraits):
    """ Defines a company with departments and employees. """

    name = Str('<unknown>')
    departments = List(Department)
    employees = List(Employee)


# Create an empty view for objects that have no data to display:
no_view = View()

# Define the TreeEditor used to display the hierarchy:
tree_editor = TreeEditor(
    nodes=[
        # The first node specified is the top level one
        TreeNode(
            node_for=[Company],
            auto_open=True,
            # child nodes are
            children='',
            label='name',  # label with Company name
            view=View(['name'])
        ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='departments',
            label='=Departments',  # constant label
            view=no_view,
            add=[Department],
         ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='employees',
            label='=Employees',   # constant label
            view=no_view,
            add=[Employee]
        ),
        TreeNode(
            node_for=[Department],
            auto_open=True,
            children='employees',
            label='name',   # label with Department name
            view=View(['name']),
            add=[Employee]
        ),
        TreeNode(
            node_for=[Employee],
            auto_open=True,
            label='name',   # label with Employee name
            view=View(['name', 'title', 'phone'])
        )
    ]
)


class Partner(HasTraits):
    """ Defines a business partner."""

    name = Str('<unknown>')
    company = Instance(Company)

    traits_view = View(
        Item(name='company', editor=tree_editor, show_label=False),
        title='Company Structure',
        buttons=['OK'],
        resizable=True,
        style='custom',
        width=.3,
        height=500
    )


# Create an example data structure:
jason = Employee(name='Jason', title='Senior Engineer', phone='536-1057')
mike = Employee(name='Mike', title='Senior Engineer', phone='536-1057')
dave = Employee(name='Dave', title='Senior Developer', phone='536-1057')
martin = Employee(name='Martin', title='Senior Engineer', phone='536-1057')
duncan = Employee(name='Duncan', title='Consultant', phone='526-1057')

# Create the demo:
demo = Partner(
    name='Enthought, Inc.',
    company=Company(
        name='Enthought',
        employees=[dave, martin, duncan, jason, mike],
        departments=[
            Department(
                name='Business',
                employees=[jason, mike]
            ),
            Department(
                name='Scientific',
                employees=[dave, martin, duncan]
            )
        ]
    )
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# enum_editor.py -- Example of using an enumeration editor

from traits.api import Enum, HasTraits
from traitsui.api import EnumEditor, Item, View


class EnumExample(HasTraits):
    priority = Enum('Medium', 'Highest', 'High', 'Low', 'Lowest')

    view = View(
        Item(
            name='priority',
            editor=EnumEditor(
                values={
                    'Highest': '1:Highest',
                    'High': '2:High',
                    'Medium': '3:Medium',
                    'Low': '4:Low',
                    'Lowest': '5:Lowest',
                }
            ),
        ),
    )


EnumExample().configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the tree-based Python value editor and the value editor factory.
"""

from traits.api import Bool, Instance, Int

from traitsui.editor import Editor
from traitsui.editor_factory import EditorFactory
from traitsui.editors.tree_editor import TreeEditor
from traitsui.item import Item
from traitsui.value_tree import RootNode, value_tree_nodes
from traitsui.view import View

# -------------------------------------------------------------------------
#  'SimpleEditor' class:
# -------------------------------------------------------------------------


class _ValueEditor(Editor):
    """ Simple style of editor for values, which displays a tree.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the editor read only?
    readonly = Bool(False)

    #: The root node of the value tree
    root = Instance(RootNode)

    #: Is the value editor scrollable? This values overrides the default.
    scrollable = True

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.update_editor()
        editor = TreeEditor(
            auto_open=self.factory.auto_open,
            hide_root=True,
            editable=False,
            nodes=value_tree_nodes,
        )
        self._ui = self.edit_traits(
            parent=parent,
            view=View(
                Item("root", show_label=False, editor=editor), kind="subpanel"
            ),
        )
        self._ui.parent = self.ui
        self.control = self._ui.control

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        self.root = RootNode(name="", value=self.value, readonly=self.readonly)

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        self._ui.dispose()

        super().dispose()

    def get_error_control(self):
        """ Returns the editor's control for indicating error status.
        """
        return self._ui.get_error_controls()


class ValueEditor(EditorFactory):
    """ Editor factory for tree-based value editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Number of tree levels to automatically open
    auto_open = Int(2)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = ValueEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines tree node classes and editors for various types of values.
"""

import inspect
from operator import itemgetter

from types import FunctionType, MethodType

from traits.api import (
    Any,
    Bool,
    HasPrivateTraits,
    HasTraits,
    Instance,
    List,
    Str,
)

from .tree_node import ObjectTreeNode, TreeNode, TreeNodeObject

from .editors.tree_editor import TreeEditor


class SingleValueTreeNodeObject(TreeNodeObject):
    """ A tree node for objects of types that have a single value.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The parent of this node
    parent = Instance(TreeNodeObject)

    #: Name of the value
    name = Str()

    #: User-specified override of the default label
    label = Str()

    #: The value itself
    value = Any()

    #: Is the value readonly?
    readonly = Bool(False)

    def tno_allows_children(self, node):
        """ Returns whether this object can have children (False for this
        class).
        """
        return False

    def tno_has_children(self, node):
        """ Returns whether the object has children (False for this class).
        """
        return False

    def tno_can_rename(self, node):
        """ Returns whether the object's children can be renamed (False for
        this class).
        """
        return False

    def tno_can_copy(self, node):
        """ Returns whether the object's children can be copied (True for this
        class).
        """
        return True

    def tno_can_delete(self, node):
        """ Returns whether the object's children can be deleted (False for
        this class).
        """
        return False

    def tno_can_insert(self, node):
        """ Returns whether the object's children can be inserted (False,
        meaning children are appended, for this class).
        """
        return False

    def tno_get_icon(self, node, is_expanded):
        """ Returns the icon for a specified object.
        """
        return "@icons:%s_node" % self.__class__.__name__[:-4].lower()

    def tno_set_label(self, node, label):
        """ Sets the label for a specified object.
        """
        if label == "?":
            label = ""
        self.label = label

    def tno_get_label(self, node):
        """ Gets the label to display for a specified object.
        """
        if self.label != "":
            return self.label

        if self.name == "":
            return self.format_value(self.value)

        return "%s: %s" % (self.name, self.format_value(self.value))

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return repr(value)

    def node_for(self, name, value):
        """ Returns the correct node type for a specified value.
        """
        for type, node in basic_types():
            if isinstance(value, type):
                break
        else:
            node = OtherNode
            if inspect.isclass(value):
                node = ClassNode

            elif hasattr(value, "__class__"):
                node = ObjectNode

        return node(
            parent=self, name=name, value=value, readonly=self.readonly
        )


class MultiValueTreeNodeObject(SingleValueTreeNodeObject):
    """ A tree node for objects of types that have multiple values.
    """

    def tno_allows_children(self, node):
        """ Returns whether this object can have children (True for this class).
        """
        return True

    def tno_has_children(self, node):
        """ Returns whether the object has children (True for this class).
        """
        return True


class StringNode(SingleValueTreeNodeObject):
    """ A tree node for strings.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        n = len(value)
        if len(value) > 80:
            value = "%s...%s" % (value[:42], value[-35:])

        return "%s [%d]" % (repr(value), n)


class NoneNode(SingleValueTreeNodeObject):
    """ A tree node for None values.
    """

    pass


class BoolNode(SingleValueTreeNodeObject):
    """ A tree node for Boolean values.
    """

    pass


class IntNode(SingleValueTreeNodeObject):
    """ A tree node for integer values.
    """

    pass


class FloatNode(SingleValueTreeNodeObject):
    """ A tree node for floating point values.
    """

    pass


class ComplexNode(SingleValueTreeNodeObject):
    """ A tree node for complex number values.
    """

    pass


class OtherNode(SingleValueTreeNodeObject):
    """ A tree node for single-value types for which there is not another
    node type.
    """

    pass


class TupleNode(MultiValueTreeNodeObject):
    """ A tree node for tuples.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "Tuple(%d)" % len(value)

    def tno_has_children(self, node):
        """ Returns whether the object has children, based on the length of
            the tuple.
        """
        return len(self.value) > 0

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        node_for = self.node_for
        value = self.value
        if len(value) > 500:
            return (
                [node_for("[%d]" % i, x) for i, x in enumerate(value[:250])]
                + [StringNode(value="...", readonly=True)]
                + [node_for("[%d]" % i, x) for i, x in enumerate(value[-250:])]
            )

        return [node_for("[%d]" % i, x) for i, x in enumerate(value)]


class ListNode(TupleNode):
    """ A tree node for lists.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "List(%d)" % len(value)

    def tno_can_delete(self, node):
        """ Returns whether the object's children can be deleted.
        """
        return not self.readonly

    def tno_can_insert(self, node):
        """ Returns whether the object's children can be inserted (vs.
        appended).
        """
        return not self.readonly


class SetNode(ListNode):
    """ A tree node for sets.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "Set(%d)" % len(value)


class ArrayNode(TupleNode):
    """ A tree node for arrays.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "Array(%s)" % ",".join([str(n) for n in value.shape])


class DictNode(TupleNode):
    """ A tree node for dictionaries.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "Dict(%d)" % len(value)

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        node_for = self.node_for
        items = [(repr(k), v) for k, v in self.value.items()]
        items.sort(key=itemgetter(0))
        if len(items) > 500:
            return (
                [node_for("[%s]" % k, v) for k, v in items[:250]]
                + [StringNode(value="...", readonly=True)]
                + [node_for("[%s]" % k, v) for k, v in items[-250:]]
            )

        return [node_for("[%s]" % k, v) for k, v in items]

    def tno_can_delete(self, node):
        """ Returns whether the object's children can be deleted.
        """
        return not self.readonly


class FunctionNode(SingleValueTreeNodeObject):
    """ A tree node for functions
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return "Function %s()" % (value.__name__)


# ---------------------------------------------------------------------------
#  'MethodNode' class:
# ---------------------------------------------------------------------------


class MethodNode(MultiValueTreeNodeObject):
    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        type = "B"
        if value.__self__ is None:
            type = "Unb"

        return "%sound method %s.%s()" % (
            type,
            value.__self__.__class__.__name__,
            value.__func__.__name__,
        )

    def tno_has_children(self, node):
        """ Returns whether the object has children.
        """
        return self.value.__func__ is not None

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        return [self.node_for("Object", self.value.__self__)]


class ObjectNode(MultiValueTreeNodeObject):
    """ A tree node for objects.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        try:
            klass = value.__class__.__name__
        except:
            klass = "???"
        return "%s(0x%08X)" % (klass, id(value))

    def tno_has_children(self, node):
        """ Returns whether the object has children.
        """
        try:
            return len(self.value.__dict__) > 0
        except:
            return False

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        items = [(k, v) for k, v in self.value.__dict__.items()]
        items.sort(key=itemgetter(0))
        return [self.node_for("." + k, v) for k, v in items]


class ClassNode(ObjectNode):
    """ A tree node for classes.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return value.__name__


class TraitsNode(ObjectNode):
    """ A tree node for traits.
    """

    def tno_has_children(self, node):
        """ Returns whether the object has children.
        """
        return len(self._get_names()) > 0

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        names = sorted(self._get_names())
        value = self.value
        node_for = self.node_for
        nodes = []
        for name in names:
            try:
                item_value = getattr(value, name, "<unknown>")
            except Exception as excp:
                item_value = "<%s>" % excp
            nodes.append(node_for("." + name, item_value))

        return nodes

    def _get_names(self):
        """ Gets the names of all defined traits or attributes.
        """
        value = self.value
        names = {}
        for name in value.trait_names(type=lambda x: x != "event"):
            names[name] = None
        for name in value.__dict__.keys():
            names[name] = None
        return list(names.keys())

    def tno_when_children_replaced(self, node, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
        specified object.
        """
        self._listener = listener
        self.value.on_trait_change(
            self._children_replaced, remove=remove, dispatch="ui"
        )

    def _children_replaced(self):
        self._listener(self)

    def tno_when_children_changed(self, node, listener, remove):
        """ Sets up or removes a listener for children being changed on a
        specified object.
        """
        pass


class RootNode(MultiValueTreeNodeObject):
    """ A root node.
    """

    def format_value(self, value):
        """ Returns the formatted version of the value.
        """
        return ""

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        return [self.node_for("", self.value)]


# -------------------------------------------------------------------------
#  Define the mapping of object types to nodes:
# -------------------------------------------------------------------------

_basic_types = None


def basic_types():
    global _basic_types

    if _basic_types is None:
        # Create the mapping of object types to nodes:
        _basic_types = [
            (type(None), NoneNode),
            (str, StringNode),
            (str, StringNode),
            (bool, BoolNode),
            (int, IntNode),
            (float, FloatNode),
            (complex, ComplexNode),
            (tuple, TupleNode),
            (list, ListNode),
            (set, SetNode),
            (dict, DictNode),
            (FunctionType, FunctionNode),
            (MethodType, MethodNode),
            (HasTraits, TraitsNode),
        ]

        try:
            from numpy import array

            _basic_types.append((type(array([1])), ArrayNode))
        except ImportError:
            pass

    return _basic_types


class _ValueTree(HasPrivateTraits):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: List of arbitrary Python values contained in the tree:
    values = List(SingleValueTreeNodeObject)


# -------------------------------------------------------------------------
#  Defines the value tree editor(s):
# -------------------------------------------------------------------------

# Nodes in a value tree:
value_tree_nodes = [
    ObjectTreeNode(
        node_for=[
            NoneNode,
            StringNode,
            BoolNode,
            IntNode,
            FloatNode,
            ComplexNode,
            OtherNode,
            TupleNode,
            ListNode,
            ArrayNode,
            DictNode,
            SetNode,
            FunctionNode,
            MethodNode,
            ObjectNode,
            TraitsNode,
            RootNode,
            ClassNode,
        ]
    )
]

# Editor for a value tree:
value_tree_editor = TreeEditor(
    auto_open=3, hide_root=True, editable=False, nodes=value_tree_nodes
)

# Editor for a value tree with a root:
value_tree_editor_with_root = TreeEditor(
    auto_open=3,
    editable=False,
    nodes=[
        ObjectTreeNode(
            node_for=[
                NoneNode,
                StringNode,
                BoolNode,
                IntNode,
                FloatNode,
                ComplexNode,
                OtherNode,
                TupleNode,
                ListNode,
                ArrayNode,
                DictNode,
                SetNode,
                FunctionNode,
                MethodNode,
                ObjectNode,
                TraitsNode,
                RootNode,
                ClassNode,
            ]
        ),
        TreeNode(
            node_for=[_ValueTree],
            auto_open=True,
            children="values",
            move=[SingleValueTreeNodeObject],
            copy=False,
            label="=Values",
            icon_group="traits_node",
            icon_open="traits_node",
        ),
    ],
)

# -------------------------------------------------------------------------
#  Defines a 'ValueTree' trait:
# -------------------------------------------------------------------------

# Trait for a value tree:
ValueTree = Instance(_ValueTree, (), editor=value_tree_editor_with_root)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Define the top-level object which is responsible for dispatching testing
functionality for a given GUI object. The functionality is exposed via
``UITester``, which is a TraitsUI specific tester.
"""

from contextlib import contextmanager
import textwrap

from traitsui.testing._exception_handling import (
    reraise_exceptions as _reraise_exceptions,
)
from traitsui.testing._gui import (
    process_cascade_events as _process_cascade_events,
)
from traitsui.testing.tester import locator
from traitsui.testing.tester.exceptions import (
    InteractionNotSupported,
    LocationNotSupported,
)


class UIWrapper:
    """
    An ``UIWrapper`` has the following responsibilities:

    (1) Wraps a UI target.
    (2) Search for a nested UI target within the wrapped UI target.
    (3) Perform user interaction on the UI target, e.g. mouse click.

    A UI target is an object which can be searched for nested UI targets,
    either as intermediate things (like editors or a table widget), or as a
    leaf widget which can be operated upon (e.g. a mouse click).

    For example, a ``UIWrapper`` may wrap an instance of traitsui.ui.UI in
    which more UI targets can be located. A ``UIWrapper`` may also wrap a
    leaf widget on which user interactions can be performed.

    For locating a nested UI target, the ``locate`` method is provided.
    For simulating user interactions such as a mouse click or visual
    inspection, the ``perform`` and ``inspect`` methods are provided.

    Parameters
    ----------
    target : any
        An object on which further UI target can be searched for, or can be
        a leaf target that can be operated on.
    registries : list of AbstractTargetRegistry
        Registries of interaction for different target, in the order
        of decreasing priority.
    delay : int, optional
        Time delay (in ms) in which actions by the wrapper are performed. Note
        it is propagated through to created child wrappers. The delay allows
        visual confirmation test code is working as desired. Defaults to 0.
    auto_process_events : bool, optional
        Whether to process (cascade) GUI events automatically. Default is True.
        For tests that launch a modal dialog and rely on a recurring timer to
        poll if the dialog is closed, it may be necessary to set this flag to
        false in order to avoid deadlocks. Note that this is propagated
        through to created child wrappers.

    Attributes
    ----------
    delay : int
        Time delay (in ms) in which actions by the wrapper are performed. Note
        it is propagated through to created child wrappers. The delay allows
        visual confirmation test code is working as desired.
    _target : any
        The UI target currently wrapped for interactions or locating nested
        components. This is a protected attribute intended to be used for
        extending the testing API. Usage of this attribute may expose the
        software's internal structure to the tests, developers should use this
        attribute with discretion based on context.
    """

    def __init__(
            self, target, *, registries, delay=0, auto_process_events=True):
        self._target = target
        self._registries = registries
        self._auto_process_events = auto_process_events
        self.delay = delay

    def help(self):
        """ Print help messages.
        (This function is intended for interactive use.)
        """
        # mapping from interaction types to their documentation
        interaction_to_doc = dict()

        # mapping from location types to their documentation
        location_to_doc = dict()

        # Order registries by their priority (low to high)
        for registry in self._registries[::-1]:
            for type_ in registry._get_interactions(self._target):
                interaction_to_doc[type_] = registry._get_interaction_doc(
                    target=self._target,
                    interaction_class=type_,
                )

            for type_ in registry._get_locations(self._target):
                location_to_doc[type_] = registry._get_location_doc(
                    target=self._target,
                    locator_class=type_,
                )

        print("Interactions")
        print("------------")
        for interaction_type in sorted(interaction_to_doc, key=repr):
            print(repr(interaction_type))
            print(
                textwrap.indent(
                    interaction_to_doc[interaction_type], prefix="    "
                )
            )
            print()

        if not interaction_to_doc:
            print("No interactions are supported.")
            print()

        print("Locations")
        print("---------")
        for locator_type in sorted(location_to_doc, key=repr):
            print(repr(locator_type))
            print(
                textwrap.indent(
                    location_to_doc[locator_type], prefix="    "
                )
            )
            print()

        if not location_to_doc:
            print("No locations are supported.")
            print()

    def locate(self, location):
        """ Attempt to resolve the given location and return a new
        UIWrapper.

        Parameters
        ----------
        location : any
            An instance of a location type supported by the current target.
            e.g. ``traitsui.testing.api.Index(1)``

        Returns
        -------
        wrapper : UIWrapper
            A new UIWrapper for the given location.

        Raises
        ------
        LocationNotSupported
            If the given location is not supported.

        See Also
        --------
        UIWrapper.help
        """
        return UIWrapper(
            target=self._get_next_target(location),
            registries=self._registries,
            delay=self.delay,
            auto_process_events=self._auto_process_events,
        )

    def find_by_name(self, name):
        """ Find a target inside the current target using a name.

        This is equivalent to calling ``locate(TargetByName(name=name))``.

        Parameters
        ----------
        name : str
            A single name for retreiving a target on a UI.

        Returns
        -------
        wrapper : UIWrapper

        Raises
        ------
        LocationNotSupported
            If the current target does not support locating another target
            by a name.

        See Also
        --------
        traitsui.testing.tester.locator.TargetByName
        """
        return self.locate(locator.TargetByName(name=name))

    def find_by_id(self, id):
        """ Find a target inside the current target using an id.

        This is equivalent to calling ``locate(TargetById(id=id))``.

        Parameters
        ----------
        id : str
            Id for finding an item in the UI.

        Returns
        -------
        wrapper : UIWrapper

        Raises
        ------
        LocationNotSupported
            If the current target does not support locating another target
            by a unique identifier.

        See Also
        --------
        traitsui.testing.tester.locator.TargetById
        """
        return self.locate(locator.TargetById(id=id))

    def perform(self, interaction):
        """ Perform a user interaction that causes side effects.

        Parameters
        ----------
        interaction : object
            An instance of an interaction type supported by the current target.
            e.g. ``traitsui.testing.api.MouseClick()``

        Raises
        ------
        InteractionNotSupported
            If the interaction given is not supported for the current UI
            target.

        See Also
        --------
        UIWrapper.help
        """
        self._perform_or_inspect(interaction)

    def inspect(self, interaction):
        """ Return a value or values for inspection.

        Parameters
        ----------
        interaction : object
            An instance of an interaction type supported by the current target.
            e.g. ``traitsui.testing.api.DisplayedText()``

        Returns
        -------
        values : any
            Any values returned for the given inspection. The type should be
            documented by the interaction type object.

        Raises
        ------
        InteractionNotSupported
            If the interaction given is not supported for the current UI
            target.

        See Also
        --------
        UIWrapper.help
        """
        return self._perform_or_inspect(interaction)

    # Private methods #########################################################

    def _perform_or_inspect(self, interaction):
        """ Perform a user interaction or a user inspection.

        Parameters
        ----------
        interaction : instance of interaction type
            An object defining the interaction.

        Returns
        -------
        value : any

        Raises
        ------
        InteractionNotSupported
            If the given interaction does not have a corresponding
            implementation for the wrapped UI target.
        """
        supported = []
        for registry in self._registries:
            try:
                handler = registry._get_handler(
                    target=self._target,
                    interaction=interaction,
                )
            except InteractionNotSupported as e:
                supported.extend(e.supported)
                continue
            else:
                context = (
                    _event_processed if self._auto_process_events
                    else _nullcontext
                )
                with context():
                    return handler(self, interaction)

        raise InteractionNotSupported(
            target_class=self._target.__class__,
            interaction_class=interaction.__class__,
            supported=supported,
        )

    def _get_next_target(self, location):
        """ Return the next UI target from the given location.

        Parameters
        ----------
        location : instance of locator type
            A location for resolving the next target.

        Returns
        -------
        new_target : any

        Raises
        ------
        LocationNotSupport
            If no solver are provided for resolving the given location in the
            wrapped UI target.
        """
        supported = set()
        for registry in self._registries:
            try:
                handler = registry._get_solver(
                    target=self._target,
                    location=location,
                )
            except LocationNotSupported as e:
                supported |= set(e.supported)
            else:
                if self._auto_process_events:
                    with _reraise_exceptions():
                        _process_cascade_events()
                return handler(self, location)

        raise LocationNotSupported(
            target_class=self._target.__class__,
            locator_class=location.__class__,
            supported=list(supported),
        )


@contextmanager
def _event_processed():
    """ Context manager to ensure GUI events are processed upon entering
    and exiting the context.
    """
    with _reraise_exceptions():
        _process_cascade_events()
        try:
            yield
        finally:
            _process_cascade_events()


@contextmanager
def _nullcontext():
    """ Equivalent to contextlib.nullcontext() in Python >= 3.7"""
    yield



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# tree_editor.py -- Example of a tree editor

from traits.api import HasTraits, Instance, List, Str, Regex
from traitsui.api import (
    Action, Group, Handler, HGroup, Item, Menu, Separator, TreeEditor,
    TreeNode, View, VSplit,
)
from traitsui.editors.tree_editor import (
    NewAction, CopyAction, CutAction, PasteAction, DeleteAction, RenameAction,
)


class Employee(HasTraits):
    name = Str('<unknown>')
    title = Str()
    phone = Regex(regex=r'\d\d\d-\d\d\d\d')

    def default_title(self):
        self.title = 'Senior Engineer'


class Department(HasTraits):
    name = Str('<unknown>')
    employees = List(Employee)


class Company(HasTraits):
    name = Str('<unknown>')
    departments = List(Department)
    employees = List(Employee)


class Owner(HasTraits):
    name = Str('<unknown>')
    company = Instance(Company)


jason = Employee(name='Jason', title='Engineer', phone='536-1057')
mike = Employee(name='Mike', title='Sr. Marketing Analyst', phone='536-1057')
dave = Employee(name='Dave', title='Sr. Engineer', phone='536-1057')
susan = Employee(name='Susan', title='Engineer', phone='536-1057')
betty = Employee(name='Betty', title='Marketing Analyst')

owner = Owner(
    name='wile',
    company=Company(
        name='Acme Labs, Inc.',
        departments=[
            Department(name='Marketing', employees=[mike, betty]),
            Department(name='Engineering', employees=[dave, susan, jason]),
        ],
        employees=[dave, susan, mike, betty, jason],
    ),
)

# View for objects that aren't edited
no_view = View()

# Actions used by tree editor context menu
def_title_action = Action(name='Default title', action='object.default')

dept_action = Action(
    name='Department',
    action='handler.employee_department(editor,object)',
)

# View used by tree editor
employee_view = View(
    VSplit(
        HGroup('3', 'name'),
        HGroup('9', 'title'),
        HGroup('phone'),
        id='vsplit',
    ),
    id='traits.doc.example.treeeditor',
    dock='vertical',
)


class TreeHandler(Handler):

    def employee_department(self, editor, object):
        dept = editor.get_parent(object)
        print(f'{object.name} works in the {dept.name} department.')


tree_editor = TreeEditor(
    nodes=[
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='',
            label='name',
            view=View(Group('name', orientation='vertical', show_left=True)),
        ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='departments',
            label='=Departments',
            view=no_view,
            add=[Department],
        ),
        TreeNode(
            node_for=[Company],
            auto_open=True,
            children='employees',
            label='=Employees',
            view=no_view,
            add=[Employee],
        ),
        TreeNode(
            node_for=[Department],
            auto_open=True,
            children='employees',
            label='name',
            menu=Menu(
                NewAction,
                Separator(),
                DeleteAction,
                Separator(),
                RenameAction,
                Separator(),
                CopyAction,
                CutAction,
                PasteAction,
            ),
            view=View(Group('name', orientation='vertical', show_left=True)),
            add=[Employee],
        ),
        TreeNode(
            node_for=[Employee],
            auto_open=True,
            label='name',
            menu=Menu(
                NewAction,
                Separator(),
                def_title_action,
                dept_action,
                Separator(),
                CopyAction,
                CutAction,
                PasteAction,
                Separator(),
                DeleteAction,
                Separator(),
                RenameAction,
            ),
            view=employee_view,
        ),
    ],
)

# The main view
view = View(
    Group(
        Item(name='company', id='company', editor=tree_editor, resizable=True),
        orientation='vertical',
        show_labels=True,
        show_left=True,
    ),
    title='Company Structure',
    id='traitsui.tests.tree_editor_test',
    dock='horizontal',
    drop_class=HasTraits,
    handler=TreeHandler(),
    buttons=['Undo', 'OK', 'Cancel'],
    resizable=True,
    width=.3,
    height=.3,
)

if __name__ == '__main__':
    owner.configure_traits(view=view)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import pickle
import unittest
from unittest.mock import patch

from traitsui.tests._tools import BaseTestMixin, requires_toolkit, ToolkitName
from traitsui.theme import Theme


class TestTheme(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)
        self.theme = Theme()

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_theme_pickling(self):
        # A regression test for issue enthought/traitsui#825
        pickle.dumps(self.theme)

    @requires_toolkit([ToolkitName.wx])
    def test_theme_content_color_default(self):
        import wx
        self.assertEqual(self.theme.content_color, wx.BLACK)

    def test_theme_content_color_setter_getter(self):
        self.theme.content_color = "red"
        self.assertEqual(self.theme.content_color, "red")

    @requires_toolkit([ToolkitName.wx])
    def test_theme_label_color_default(self):
        import wx
        self.assertEqual(self.theme.label_color, wx.BLACK)

    def test_theme_label_color_setter_getter(self):
        self.theme.label_color = "red"
        self.assertEqual(self.theme.label_color, "red")

    @requires_toolkit([ToolkitName.wx])
    def test_theme_get_image_slice(self):
        self.theme.image = "mock_image"
        with patch("traitsui.wx.image_slice.image_slice_for") as slice_func:
            self.theme.image_slice
            slice_func.assert_called_once()

    def test_theme_get_image_slice_none(self):
        self.theme.image = None
        self.assertIsNone(self.theme.image_slice)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the color editor factory for the all traits toolkit backends.
"""

from traits.api import Bool

from traitsui.editor_factory import EditorFactory
from traitsui.toolkit import toolkit_object
from traitsui.view import View

# -------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
# -------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for color editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Is the underlying color trait mapped?
    mapped = Bool(True)

    #: Do we use a native dialog for the popup or the toolkit's?
    #: At present, only affects Qt.
    use_native_dialog = Bool(True)

    # -------------------------------------------------------------------------
    #  Traits view definition:
    # -------------------------------------------------------------------------

    traits_view = View(["mapped{Is value mapped?}", "|[]>"])


# Define the ColorEditor class
# The function will try to return the toolkit-specific editor factory (located
# in traitsui.<toolkit>.color_editor, and if none is found, the
# ToolkitEditorFactory declared here is returned.
def ColorEditor(*args, **traits):
    r""" Returns an instance of the toolkit-specific editor factory declared in
    traitsui.<toolkit>.color_editor. If such an editor factory
    cannot be located, an instance of the abstract ToolkitEditorFactory
    declared in traitsui.editors.color_editor is returned.

    Parameters
    ----------
    \*args, \*\*traits
        arguments and keywords to be passed on to the editor
        factory's constructor.
    """

    try:
        return toolkit_object("color_editor:ToolkitEditorFactory", True)(
            *args, **traits
        )
    except:
        return ToolkitEditorFactory(*args, **traits)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import wx

from traitsui.wx.check_list_editor import CustomEditor
from traitsui.testing.tester.command import MouseClick
from traitsui.testing.tester.locator import Index
from traitsui.testing.tester._ui_tester_registry._common_ui_targets import (
    BaseSourceWithLocation
)
from traitsui.testing.tester._ui_tester_registry._layout import (
    column_major_to_row_major
)
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers


class _IndexedCustomCheckListEditor(BaseSourceWithLocation):
    """ Wrapper for CheckListEditor + Index
    """
    source_class = CustomEditor
    locator_class = Index
    handlers = [
        (MouseClick,
            (lambda wrapper, _:
                _interaction_helpers.mouse_click_checkbox_child_in_panel(
                    control=wrapper._target.source.control,
                    index=convert_index(
                        source=wrapper._target.source,
                        index=wrapper._target.location.index
                    ),
                    delay=wrapper.delay))),
    ]


def convert_index(source, index):
    """ Helper function to convert an index for a GridSizer so that the
    index counts over the grid in the correct direction.
    The grid is always populated in row major order, however, the elements
    are assigned to each entry in the grid so that when displayed they appear
    in column major order.
    Sizers are indexed in the order they are populated, so to access
    the correct element we may need to convert a column-major based index
    into a row-major one.

    Parameters
    ----------
    control : CustomEditor
        The Custom CheckList Editor of interest. Its control is the wx.Panel
        containing child objects organized with a wx.GridSizer
    index : int
        the index of interest
    """
    sizer = source.control.GetSizer()
    if isinstance(sizer, wx.BoxSizer):
        return index
    n = len(source.names)
    num_cols = sizer.GetCols()
    num_rows = sizer.GetEffectiveRowsCount()
    return column_major_to_row_major(index, n, num_rows, num_cols)


def register(registry):
    """ Register interactions for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    _IndexedCustomCheckListEditor.register(registry)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the key binding editor for use with the KeyBinding class. This is a
specialized editor used to associate a particular key with a control (i.e., the
key binding editor).
"""

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object

# Callable which returns the editor to use in the ui.


def key_binding_editor(*args, **traits):
    return toolkit_object("key_binding_editor:KeyBindingEditor")(
        *args, **traits
    )


# -------------------------------------------------------------------------
#  Create the editor factory object:
# -------------------------------------------------------------------------
KeyBindingEditor = ToolkitEditorFactory = BasicEditorFactory(
    klass=key_binding_editor
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import wx

from traitsui.wx.button_editor import SimpleEditor, CustomEditor
from traitsui.testing.tester.command import MouseClick
from traitsui.testing.tester.query import DisplayedText
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers


def mouse_click_ImageButton(wrapper, interaction):
    """ Performs a mouce click on an pyface.ui.wx.ImageButton object.

    Parameters
    ----------
    wrapper : UIWrapper
        The wrapper object wrapping the Custom Button Editor which utilizes
        an ImageButton.
    interaction : instance of traitsui.testing.tester.command.MouseClick
        interaction is unused here, but it is included so that the function
        matches the expected format for a handler.  Note this handler is
        intended to be used with an interaction_class of a MouseClick.
    """

    control = wrapper._target.control
    if not control.IsEnabled():
        return
    wx.MilliSleep(wrapper.delay)

    left_down_event = wx.MouseEvent(
        wx.wxEVT_LEFT_DOWN
    )
    left_up_event = wx.MouseEvent(
        wx.wxEVT_LEFT_UP
    )
    control.ProcessEvent(left_down_event)
    control.ProcessEvent(left_up_event)


def register(registry):
    """ Register solvers/handlers specific to wx Button Editors
    for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
    """

    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=MouseClick,
        handler=(lambda wrapper, _: _interaction_helpers.mouse_click_button(
                 control=wrapper._target.control, delay=wrapper.delay))
    )

    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=DisplayedText,
        handler=lambda wrapper, _: wrapper._target.control.GetLabel()
    )

    registry.register_interaction(
        target_class=CustomEditor,
        interaction_class=MouseClick,
        handler=mouse_click_ImageButton
    )

    registry.register_interaction(
        target_class=CustomEditor,
        interaction_class=DisplayedText,
        handler=lambda wrapper, _: wrapper._target.control.GetLabel()
    )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the manager for Undo and Redo history for Traits user interface
    support.
"""

import collections.abc

from traits.api import (
    Bool,
    Event,
    HasPrivateTraits,
    HasStrictTraits,
    HasTraits,
    Instance,
    Int,
    List,
    Property,
    Str,
    Trait,
    cached_property,
    observe,
)
from pyface.undo.api import (
    AbstractCommand, CommandStack, ICommand, ICommandStack, IUndoManager,
    UndoManager
)


NumericTypes = (int, float, complex)
SimpleTypes = (str, bytes) + NumericTypes


class AbstractUndoItem(AbstractCommand):
    """ Abstract base class for undo items.

    This class is deprecated and will be removed in TraitsUI 8.  Any custom
    subclasses of this class should either subclass from AbstractCommand, or
    provide the ICommand interface.
    """

    #: A simple default name.
    name = "Edit"

    def do(self):
        """ Does nothing.
        
        All undo items log events after they have happened, so by default
        they do not do anything when added to the history.
        """
        pass

    def undo(self):
        """ Undoes the change.
        """
        raise NotImplementedError

    def redo(self):
        """ Re-does the change.
        """
        raise NotImplementedError

    def merge(self, other):
        """ Merges two undo items if possible.
        """
        import warnings
        warnings.warn(
            "'merge_undo' is deprecated and will be removed in TraitsUI 8, "
            "use 'merge' instead",
            DeprecationWarning,
        )
        return self.merge_undo(other)

    def merge_undo(self, undo_item):
        """ Merges two undo items if possible.

        This method is deprecated.
        """
        return False


class UndoItem(AbstractUndoItem):
    """ A change to an object trait, which can be undone.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Object the change occurred on
    object = Instance(HasTraits)

    #: Name of the trait that changed
    name = Str()

    #: Old value of the changed trait
    old_value = Property()

    #: New value of the changed trait
    new_value = Property()

    def _get_old_value(self):
        return self._old_value

    def _set_old_value(self, value):
        if isinstance(value, list):
            value = value[:]
        self._old_value = value

    def _get_new_value(self):
        return self._new_value

    def _set_new_value(self, value):
        if isinstance(value, list):
            value = value[:]
        self._new_value = value

    def undo(self):
        """ Undoes the change.
        """
        try:
            setattr(self.object, self.name, self.old_value)
        except Exception:
            from traitsui.api import raise_to_debug

            raise_to_debug()

    def redo(self):
        """ Re-does the change.
        """
        try:
            setattr(self.object, self.name, self.new_value)
        except Exception:
            from traitsui.api import raise_to_debug

            raise_to_debug()

    def merge(self, undo_item):
        """ Merges two undo items if possible.
        """
        # Undo items are potentially mergeable only if they are of the same
        # class and refer to the same object trait, so check that first:
        if (
            isinstance(undo_item, self.__class__)
            and (self.object is undo_item.object)
            and (self.name == undo_item.name)
        ):
            v1 = self.new_value
            v2 = undo_item.new_value
            t1 = type(v1)
            if isinstance(v2, t1):

                if isinstance(v1, str):
                    # Merge two undo items if they have new values which are
                    # strings which only differ by one character (corresponding
                    # to a single character insertion, deletion or replacement
                    # operation in a text editor):
                    n1 = len(v1)
                    n2 = len(v2)
                    if abs(n1 - n2) > 1:
                        return False
                    n = min(n1, n2)
                    i = 0
                    while (i < n) and (v1[i] == v2[i]):
                        i += 1
                    if v1[i + (n2 <= n1):] == v2[i + (n2 >= n1):]:
                        self.new_value = v2
                        return True

                elif isinstance(v1, collections.abc.Sequence):
                    # Merge sequence types only if a single element has changed
                    # from the 'original' value, and the element type is a
                    # simple Python type:
                    v1 = self.old_value
                    if isinstance(v1, collections.abc.Sequence):
                        # Note: wxColour says it's a sequence type, but it
                        # doesn't support 'len', so we handle the exception
                        # just in case other classes have similar behavior:
                        try:
                            if len(v1) == len(v2):
                                diffs = 0
                                for i, item in enumerate(v1):
                                    titem = type(item)
                                    item2 = v2[i]
                                    if (
                                        (titem not in SimpleTypes)
                                        or (not isinstance(item2, titem))
                                        or (item != item2)
                                    ):
                                        diffs += 1
                                        if diffs >= 2:
                                            return False
                                if diffs == 0:
                                    return False
                                self.new_value = v2
                                return True
                        except Exception:
                            pass

                elif t1 in NumericTypes:
                    # Always merge simple numeric trait changes:
                    self.new_value = v2
                    return True
        return False

    def merge_undo(self, undo_item):
        """ Merges two undo items if possible.

        This is deprecated.
        """
        return self.merge(undo_item)

    def __repr__(self):
        """ Returns a "pretty print" form of the object.
        """
        n = self.name
        cn = self.object.__class__.__name__
        return "undo( %s.%s = %s )\nredo( %s.%s = %s )" % (
            cn,
            n,
            self.old_value,
            cn,
            n,
            self.new_value,
        )


class ListUndoItem(AbstractUndoItem):
    """ A change to a list, which can be undone.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Object that the change occurred on
    object = Instance(HasTraits)

    #: Name of the trait that changed
    name = Str()

    #: Starting index
    index = Int()

    #: Items added to the list
    added = List()

    #: Items removed from the list
    removed = List()

    def undo(self):
        """ Undoes the change.
        """
        try:
            list = getattr(self.object, self.name)
            list[self.index : (self.index + len(self.added))] = self.removed
        except Exception:
            from traitsui.api import raise_to_debug

            raise_to_debug()

    def redo(self):
        """ Re-does the change.
        """
        try:
            list = getattr(self.object, self.name)
            list[self.index : (self.index + len(self.removed))] = self.added
        except Exception:
            from traitsui.api import raise_to_debug

            raise_to_debug()

    def merge(self, undo_item):
        """ Merges two undo items if possible.
        """
        # Discard undo items that are identical to us. This is to eliminate
        # the same undo item being created by multiple listeners monitoring the
        # same list for changes:
        if (
            isinstance(undo_item, self.__class__)
            and (self.object is undo_item.object)
            and (self.name == undo_item.name)
            and (self.index == undo_item.index)
        ):
            added = undo_item.added
            removed = undo_item.removed
            if (len(self.added) == len(added)) and (
                len(self.removed) == len(removed)
            ):
                for i, item in enumerate(self.added):
                    if item is not added[i]:
                        break
                else:
                    for i, item in enumerate(self.removed):
                        if item is not removed[i]:
                            break
                    else:
                        return True
        return False

    def merge_undo(self, undo_item):
        """ Merges two undo items if possible.

        This is deprecated.
        """
        return self.merge(undo_item)

    def __repr__(self):
        """ Returns a 'pretty print' form of the object.
        """
        return "undo( %s.%s[%d:%d] = %s )" % (
            self.object.__class__.__name__,
            self.name,
            self.index,
            self.index + len(self.removed),
            self.added,
        )


class _MultiUndoItem(AbstractCommand):
    """ The _MultiUndoItem class is an internal command that unifies commands. """

    name = "Edit"

    #: The commands that make up this undo item.
    commands = List(Instance(ICommand))

    def push(self, command):
        """ Append a command, merging if possible. """
        if len(self.commands) > 0:
            merged = self.commands[-1].merge(command)
            if merged:
                return

        self.commands.append(command)

    def merge(self, other):
        """ Try and merge a command. """
        return False

    def redo(self):
        """ Redo the sub-commands. """

        for cmd in self.commands:
            cmd.redo()

    def undo(self):
        """ Undo the sub-commands. """

        for cmd in self.commands:
            cmd.undo()


class UndoHistory(HasStrictTraits):
    """ Manages a list of undoable changes.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The undo manager for the history.
    manager = Instance(IUndoManager, allow_none=False)

    #: The command stack for the history.
    stack = Instance(ICommandStack, allow_none=False)

    #: List of accumulated undo changes.  Each item is a list of
    #: AbstractUndoItems that should be done or undone as a group.
    #: This trait should be considered private.
    #: This trait is no longer used.
    history = List()

    #: The current position in the list
    now = Property(Int, observe='stack._index')

    #: Fired when state changes to undoable
    undoable = Event(False)

    #: Fired when state changes to redoable
    redoable = Event(False)

    #: Can an action be undone?
    can_undo = Property(Bool, observe='_can_undo')

    #: Can an action be redone?
    can_redo = Property(Bool, observe='_can_redo')

    _can_undo = Bool()

    _can_redo = Bool()

    def add(self, undo_item, extend=False):
        """ Adds an UndoItem to the history.
        """
        if extend:
            self.extend(undo_item)
        else:
            self.manager.active_stack = self.stack
            self.stack.push(undo_item)

    def extend(self, undo_item):
        """ Extends the undo history.

        If possible the method merges the new UndoItem with the last item in
        the history; otherwise, it appends the new item.
        """
        self.manager.active_stack = self.stack
        # get the last command in the stack
        # XXX this is using CommandStack internals that it should not
        # XXX this should be re-architected to use the macro interface
        # XXX it possibly should be removed altogether
        entries = self.stack._stack
        if len(entries) > 0:
            command = entries[-1].command
            if not isinstance(command, _MultiUndoItem):
                command = _MultiUndoItem(commands=[command])
                entries[-1].command = command
        else:
            command = _MultiUndoItem(commands=[])
        command.push(undo_item)

    def undo(self):
        """ Undoes an operation.
        """
        if self.can_undo:
            self.manager.undo()

    def redo(self):
        """ Redoes an operation.
        """
        if self.can_redo:
            self.manager.redo()

    def revert(self):
        """ Reverts all changes made so far and clears the history.
        """
        # undo everything
        self.manager.active_stack = self.stack
        self.stack.undo(sequence_nr=-1)
        self.clear()

    def clear(self):
        """ Clears the undo history.
        """
        self.manager.active_stack = self.stack
        self.stack.clear()

    @observe('manager.stack_updated')
    def _observe_stack_updated(self, event):
        """ Update undo/redo state. """
        self._can_undo = (self.manager and self.manager.undo_name != "")
        self._can_redo = (self.manager and self.manager.redo_name != "")

    @observe('_can_undo')
    def _observe_can_undo(self, event):
        self.undoable = event.new

    @observe('_can_redo')
    def _observe_can_redo(self, event):
        self.redoable = event.new

    @cached_property
    def _get_now(self):
        return self.stack._index + 1

    def _get_can_undo(self):
        return self._can_undo
        
    def _get_can_redo(self):
        return self._can_redo

    def _manager_default(self):
        manager = UndoManager()
        return manager

    def _stack_default(self):
        stack = CommandStack(undo_manager=self.manager)
        return stack


class UndoHistoryUndoItem(AbstractUndoItem):
    """ An undo item for the undo history.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The undo history to undo or redo
    history = Instance(UndoHistory)

    def undo(self):
        """ Undoes the change.
        """
        history = self.history
        for i in range(history.now - 1, -1, -1):
            items = history.history[i]
            for j in range(len(items) - 1, -1, -1):
                items[j].undo()

    def redo(self):
        """ Re-does the change.
        """
        history = self.history
        for i in range(0, history.now):
            for item in history.history[i]:
                item.redo()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# default_trait_editors.py -- Example of using default trait editors

from traits.api import Bool, HasTraits, Range, Str
from traitsui.api import Item, View


class Adult(HasTraits):
    first_name = Str()
    last_name = Str()
    age = Range(21, 99)
    registered_voter = Bool()

    traits_view = View(
        Item(name='first_name'),
        Item(name='last_name'),
        Item(name='age'),
        Item(name='registered_voter'),
    )


alice = Adult(
    first_name='Alice',
    last_name='Smith',
    age=42,
    registered_voter=True,
)

alice.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from pyface.toolkit import toolkit_object
from traits.api import HasTraits, Instance, List, Str, String
from traitsui.api import InstanceEditor, Item, View
from traitsui.tests._tools import (
    BaseTestMixin,
    requires_toolkit,
    ToolkitName,
)

from traitsui.testing.api import (
    DisplayedText,
    Index,
    IsEnabled,
    KeyClick,
    KeySequence,
    MouseClick,
    SelectedText,
    TargetByName,
    UITester
)

ModalDialogTester = toolkit_object(
    "util.modal_dialog_tester:ModalDialogTester"
)
no_modal_dialog_tester = ModalDialogTester.__name__ == "Unimplemented"


class EditedInstance(HasTraits):
    value = Str()
    traits_view = View(Item("value"), buttons=["OK"])


class NamedInstance(HasTraits):
    name = Str()
    value = Str()
    traits_view = View(Item("value"), buttons=["OK"])


class ObjectWithInstance(HasTraits):
    inst = Instance(EditedInstance, args=())


class ObjectWithList(HasTraits):
    inst_list = List(Instance(NamedInstance))
    inst = Instance(NamedInstance, args=())

    def _inst_list_default(self):
        return [
            NamedInstance(name=value, value=value)
            for value in ['one', 'two', 'three']
        ]


simple_view = View(Item("inst"), buttons=["OK"])
custom_view = View(Item("inst", style='custom'), buttons=["OK"])
selection_view = View(
    Item(
        "inst",
        editor=InstanceEditor(name='inst_list'),
        style='custom',
    ),
    buttons=["OK"],
)
modal_view = View(
    Item("inst", style="simple", editor=InstanceEditor(kind="modal"))
)


class ValidatedEditedInstance(HasTraits):
    some_string = String("A", maxlen=3)

    traits_view = View(Item('some_string'))


class ObjectWithValidatedInstance(HasTraits):
    something = Instance(ValidatedEditedInstance, args=())

    traits_view = View(
        Item(
            'something',
            editor=InstanceEditor(),
            style='custom'
        ),
        buttons=["OK", "Cancel"],
    )


class ObjectWithValidatedList(HasTraits):
    inst_list = List(Instance(HasTraits))
    inst = Instance(HasTraits, ())

    def _inst_list_default(self):
        return [
            ValidatedEditedInstance(some_string=value)
            for value in ['a', 'b', 'c']
        ]


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestInstanceEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_simple_editor(self):
        obj = ObjectWithInstance()
        tester = UITester()
        with tester.create_ui(obj, {'view': simple_view}) as ui:
            instance = tester.find_by_name(ui, "inst")
            instance.perform(MouseClick())
            value_txt = instance.find_by_name("value")
            value_txt.perform(KeySequence("abc"))
            self.assertEqual(obj.inst.value, "abc")

    def test_custom_editor(self):
        obj = ObjectWithInstance()
        tester = UITester()
        with tester.create_ui(obj, {'view': custom_view}) as ui:
            value_txt = tester.find_by_name(ui, "inst").find_by_name("value")
            value_txt.perform(KeySequence("abc"))
            self.assertEqual(obj.inst.value, "abc")

    def test_custom_editor_with_selection(self):
        obj = ObjectWithList()
        tester = UITester()
        with tester.create_ui(obj, {'view': selection_view}) as ui:
            # test that the current object is None
            self.assertIsNone(obj.inst)

            # test that the displayed text is correct
            instance = tester.find_by_name(ui, "inst")
            text = instance.inspect(SelectedText())
            self.assertEqual(text, obj.inst_list[0].name)

            # test that changing selection works
            instance.locate(Index(1)).perform(MouseClick())
            self.assertIs(obj.inst, obj.inst_list[1])

            # test that the displayed text is correct
            text = instance.inspect(SelectedText())
            self.assertEqual(text, obj.inst_list[1].name)

            # test editing the view works
            value_txt = instance.find_by_name("value")
            value_txt.perform(KeySequence("abc"))
            self.assertEqual(obj.inst.value, "twoabc")

    def test_custom_editor_with_selection_initialized(self):
        obj = ObjectWithList()
        obj.inst = obj.inst_list[1]
        tester = UITester()
        with tester.create_ui(obj, {'view': selection_view}) as ui:
            # test that the current object is the correct one
            self.assertIs(obj.inst, obj.inst_list[1])

            # test that the displayed text is correct
            instance = tester.find_by_name(ui, "inst")
            text = instance.inspect(SelectedText())
            self.assertEqual(text, obj.inst.name)

    def test_custom_editor_resynch_editor(self):
        edited_inst = EditedInstance(value='hello')
        obj = ObjectWithInstance(inst=edited_inst)
        tester = UITester()
        with tester.create_ui(obj, {'view': custom_view}) as ui:
            value_txt = tester.find_by_name(ui, "inst").find_by_name("value")
            displayed = value_txt.inspect(DisplayedText())
            self.assertEqual(displayed, "hello")
            edited_inst.value = "bye"
            displayed = value_txt.inspect(DisplayedText())
            self.assertEqual(displayed, "bye")

    def test_simple_editor_resynch_editor(self):
        edited_inst = EditedInstance(value='hello')
        obj = ObjectWithInstance(inst=edited_inst)
        tester = UITester()
        with tester.create_ui(obj, {'view': simple_view}) as ui:
            instance = tester.find_by_name(ui, "inst")
            instance.perform(MouseClick())

            value_txt = instance.find_by_name("value")
            displayed = value_txt.inspect(DisplayedText())
            self.assertEqual(displayed, "hello")
            edited_inst.value = "bye"
            displayed = value_txt.inspect(DisplayedText())
            self.assertEqual(displayed, "bye")

    def test_simple_editor_parent_closed(self):
        obj = ObjectWithInstance()
        tester = UITester()
        with tester.create_ui(obj, {'view': simple_view}) as ui:
            instance = tester.find_by_name(ui, "inst")
            instance.perform(MouseClick())

    @unittest.skipIf(no_modal_dialog_tester, "ModalDialogTester unavailable")
    def test_simple_editor_modal(self):
        # Test launching the instance editor with kind set to 'modal'
        obj = ObjectWithInstance()
        ui_tester = UITester()

        with ui_tester.create_ui(obj, dict(view=modal_view)) as ui:

            def click_button():
                ui_tester.find_by_name(ui, "inst").perform(MouseClick())

            def when_opened(tester):
                with tester.capture_error():
                    try:
                        dialog_ui = tester.get_dialog_widget()._ui
                        # If auto_process_events was not set to false, this
                        # will block due to deadlocks with ModalDialogTester.
                        ui_tester = UITester(auto_process_events=False)
                        value = ui_tester.find_by_name(dialog_ui, "value")
                        value.perform(KeySequence("Hello"))
                        self.assertEqual(obj.inst.value, "")
                        ok_button = ui_tester.find_by_id(dialog_ui, "OK")
                        ok_button.perform(MouseClick())
                    finally:
                        # If the block above fails, the dialog will block
                        # forever. Close it regardless.
                        if tester.get_dialog_widget() is not None:
                            tester.close(accept=True)

            mdtester = ModalDialogTester(click_button)
            mdtester.open_and_run(when_opened=when_opened)
            self.assertTrue(mdtester.dialog_was_opened)
            self.assertEqual(obj.inst.value, "Hello")

    # A regression test for issue enthought/traitsui#1501
    def test_propagate_errors(self):
        obj = ObjectWithValidatedInstance()
        ui_tester = UITester()
        with ui_tester.create_ui(obj) as ui:
            something_ui = ui_tester.find_by_name(ui, "something")
            some_string_field = something_ui.locate(
                TargetByName('some_string')
            )
            some_string_field.perform(KeySequence("abcd"))
            some_string_field.perform(KeyClick("Enter"))

            ok_button = ui_tester.find_by_id(ui, "OK")

            instance_editor_ui = something_ui._target._ui
            instance_editor_ui_parent = something_ui._target._ui.parent
            self.assertNotEqual(
                instance_editor_ui, ui
            )
            self.assertEqual(
                instance_editor_ui_parent, ui
            )

            self.assertEqual(
                instance_editor_ui.errors, ui.errors
            )
            self.assertFalse(ok_button.inspect(IsEnabled()))

    def test_propagate_errors_switch_selection(self):
        obj = ObjectWithValidatedList()
        ui_tester = UITester()
        with ui_tester.create_ui(obj, {'view': selection_view}) as ui:
            something_ui = ui_tester.find_by_name(ui, "inst")

            something_ui.locate(Index(0)).perform(MouseClick())

            some_string_field = something_ui.locate(
                TargetByName('some_string')
            )
            some_string_field.perform(KeySequence("bcde"))
            some_string_field.perform(KeyClick("Enter"))

            ok_button = ui_tester.find_by_id(ui, "OK")

            instance_editor_ui = something_ui._target._ui
            instance_editor_ui_parent = something_ui._target._ui.parent
            self.assertNotEqual(
                instance_editor_ui, ui
            )
            self.assertEqual(
                instance_editor_ui_parent, ui
            )

            self.assertEqual(
                instance_editor_ui.errors, ui.errors
            )
            self.assertFalse(ok_button.inspect(IsEnabled()))

            # change to a different selected that is not in an error state
            something_ui.locate(Index(1)).perform(MouseClick())

            self.assertTrue(ok_button.inspect(IsEnabled()))



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This package contains implementations for testing TraitsUI UI editors
with Qt.

The top-level module ``default_registry`` serves external packages and hides
implementation details internal to this package.
"""



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from packaging.version import Version

from traits import __version__ as TRAITS_VERSION
from traits.api import (
    HasTraits,
    Str,
)
from traits.testing.api import UnittestTools
from traitsui.api import TextEditor, View, Item
from traitsui.testing.api import (
    DisplayedText,
    KeyClick,
    KeySequence,
    MouseClick,
    InteractionNotSupported,
    UITester
)
from traitsui.tests._tools import (
    BaseTestMixin,
    GuiTestAssistant,
    no_gui_test_assistant,
    requires_toolkit,
    ToolkitName,
)


class Foo(HasTraits):

    name = Str()

    nickname = Str()


def get_view(style, auto_set):
    """ Return the default view for the Foo object.

    Parameters
    ----------
    style : str
        e.g. 'simple', or 'custom'
    auto_set : bool
        To be passed directly to the editor factory.
    """
    return View(
        Item("name", editor=TextEditor(auto_set=auto_set), style=style)
    )


# Skips tests if the backend is not either qt4 or qt5
@requires_toolkit([ToolkitName.qt])
@unittest.skipIf(no_gui_test_assistant, "No GuiTestAssistant")
class TestTextEditorQt(
        BaseTestMixin, GuiTestAssistant, UnittestTools, unittest.TestCase):
    """ Test on TextEditor with Qt backend."""

    def setUp(self):
        BaseTestMixin.setUp(self)
        GuiTestAssistant.setUp(self)

    def tearDown(self):
        GuiTestAssistant.tearDown(self)
        BaseTestMixin.tearDown(self)

    def test_text_editor_placeholder_text(self):
        foo = Foo()
        editor = TextEditor(
            placeholder="Enter name",
        )
        view = View(Item(name="name", editor=editor))

        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_editor, = ui.get_editors("name")
            self.assertEqual(
                name_editor.control.placeholderText(),
                "Enter name",
            )

    def test_text_editor_placeholder_text_and_readonly(self):
        # Placeholder can be set independently of read_only flag
        foo = Foo()
        editor = TextEditor(
            placeholder="Enter name",
            read_only=True,
        )
        view = View(Item(name="name", editor=editor))
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_editor, = ui.get_editors("name")
            self.assertEqual(
                name_editor.control.placeholderText(),
                "Enter name",
            )

    def test_text_editor_default_view(self):
        foo = Foo()
        tester = UITester()
        with tester.create_ui(foo) as ui:
            name_editor, = ui.get_editors("name")
            self.assertEqual(
                name_editor.control.placeholderText(),
                "",
            )

    def test_text_editor_custom_style_placeholder(self):
        # Test against CustomEditor using QTextEdit
        foo = Foo()
        view = View(Item(
            name="name",
            style="custom",
            editor=TextEditor(placeholder="Enter name"),
        ))
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_editor, = ui.get_editors("name")
            try:
                placeholder = name_editor.control.placeholderText()
            except AttributeError:
                # placeholderText is introduced to QTextEdit since Qt 5.2
                pass
            else:
                self.assertEqual(placeholder, "Enter name")

    def test_cancel_button(self):
        foo = Foo()
        view = View(
            Item(
                name="name",
                style="simple",
                editor=TextEditor(cancel_button=True),
            )
        )
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_editor, = ui.get_editors("name")
            # isClearButtonEnabled is introduced to QLineEdit since Qt 5.2
            if hasattr(name_editor.control, 'isClearButtonEnabled'):
                self.assertTrue(name_editor.control.isClearButtonEnabled())


# We should be able to run this test case against wx.
# Not running them now to avoid test interaction. See enthought/traitsui#752
@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestTextEditor(BaseTestMixin, unittest.TestCase, UnittestTools):
    """ Tests that can be run with any toolkit as long as there is an
    implementation for simulating user interactions.
    """
    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def check_editor_init_and_dispose(self, style, auto_set):
        # Smoke test to test setup and tear down of an editor.
        foo = Foo()
        view = get_view(style=style, auto_set=auto_set)
        with UITester().create_ui(foo, dict(view=view)):
            pass

    def test_simple_editor_init_and_dispose(self):
        # Smoke test to test setup and tear down of an editor.
        self.check_editor_init_and_dispose(style="simple", auto_set=True)

    def test_custom_editor_init_and_dispose(self):
        # Smoke test to test setup and tear down of an editor.
        self.check_editor_init_and_dispose(style="custom", auto_set=True)

    def test_readonly_editor_init_and_dispose(self):
        # Smoke test to test setup and tear down of an editor.
        self.check_editor_init_and_dispose(style="readonly", auto_set=True)

    def test_simple_editor_init_and_dispose_no_auto_set(self):
        # Smoke test to test setup and tear down of an editor.
        self.check_editor_init_and_dispose(style="simple", auto_set=False)

    def test_custom_editor_init_and_dispose_no_auto_set(self):
        # Smoke test to test setup and tear down of an editor.
        self.check_editor_init_and_dispose(style="custom", auto_set=False)

    def test_simple_auto_set_update_text(self):
        foo = Foo()
        view = get_view(style="simple", auto_set=True)
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            with self.assertTraitChanges(foo, "name", count=3):
                name_field = tester.find_by_name(ui, "name")
                name_field.perform(KeySequence("NEW"))
                # with auto-set the displayed name should match the name trait
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW")
            self.assertEqual(display_name, foo.name)

    # Currently if auto_set is false, and enter_set is false (the default
    # behavior), on Qt to ensure the text is actually set, Enter will set
    # the value
    @requires_toolkit([ToolkitName.qt])
    def test_simple_auto_set_false_do_not_update_qt(self):
        foo = Foo(name="")
        view = get_view(style="simple", auto_set=False)
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_field = tester.find_by_name(ui, "name")
            name_field.perform(KeySequence("NEW"))
            # with auto-set as False the displayed name should match what has
            # been typed not the trait itself, After "Enter" is pressed it
            # should match the name trait
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "")
            self.assertEqual(display_name, "NEW")
            name_field.perform(KeyClick("Enter"))
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW")
            self.assertEqual(display_name, foo.name)

    # If auto_set is false, the value can be set by killing the focus. This
    # can be done by simply moving to another textbox.
    @requires_toolkit([ToolkitName.wx])
    def test_simple_auto_set_false_do_not_update_wx(self):
        foo = Foo(name="")
        view = View(
                    Item("name",
                         editor=TextEditor(auto_set=False),
                         style="simple"),
                    Item("nickname",
                         editor=TextEditor(auto_set=False),
                         style="simple")
        )
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_field = tester.find_by_name(ui, "name")
            name_field.perform(KeySequence("NEW"))
            # with auto-set as False the displayed name should match what has
            # been typed not the trait itself, After moving to another textbox
            # it should match the name trait
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "")
            self.assertEqual(display_name, "NEW")
            tester.find_by_name(ui, "nickname").perform(MouseClick())
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW")
            self.assertEqual(display_name, foo.name)

    def test_custom_auto_set_true_update_text(self):
        # the auto_set flag is disregard for custom editor.  (not true on WX)
        foo = Foo()
        view = get_view(auto_set=True, style="custom")
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            with self.assertTraitChanges(foo, "name", count=3):
                name_field = tester.find_by_name(ui, "name")
                name_field.perform(KeySequence("NEW"))
            # with auto-set the displayed name should match the name trait
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW")
            self.assertEqual(display_name, foo.name)

    @requires_toolkit([ToolkitName.qt])
    def test_custom_auto_set_false_update_text(self):
        # the auto_set flag is disregard for custom editor. (not true on WX)
        foo = Foo()
        view = get_view(auto_set=False, style="custom")
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_field = tester.find_by_name(ui, "name")
            name_field.perform(KeySequence("NEW"))
            name_field.perform(KeyClick("Enter"))
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW\n")
            self.assertEqual(display_name, foo.name)

    # If auto_set is false, the value can be set by killing the focus. This
    # can be done by simply moving to another textbox.
    @requires_toolkit([ToolkitName.wx])
    def test_custom_auto_set_false_do_not_update_wx(self):
        foo = Foo(name="")
        view = View(
                    Item("name",
                         editor=TextEditor(auto_set=False),
                         style="custom"),
                    Item("nickname",
                         editor=TextEditor(auto_set=False),
                         style="custom")
        )
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_field = tester.find_by_name(ui, "name")
            name_field.perform(KeySequence("NEW"))
            # with auto-set as False the displayed name should match what has
            # been typed not the trait itself, After moving to another textbox
            # it should match the name trait
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "")
            self.assertEqual(display_name, "NEW")
            tester.find_by_name(ui, "nickname").perform(MouseClick())
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(foo.name, "NEW")
            self.assertEqual(display_name, foo.name)

    def test_readonly_editor(self):
        foo = Foo(name="A name")
        view = get_view(style="readonly", auto_set=True)
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            name_field = tester.find_by_name(ui, "name")
            with self.assertRaises(InteractionNotSupported):
                name_field.perform(KeySequence("NEW"))
            # Trying to type should do nothing
            with self.assertRaises(InteractionNotSupported):
                name_field.perform(KeyClick("Space"))
            display_name = name_field.inspect(DisplayedText())
            self.assertEqual(display_name, "A name")

    def check_format_func_used(self, style):
        # Regression test for enthought/traitsui#790
        # The test will fail with traits < 6.1.0 because the bug
        # is fixed in traits, see enthought/traitsui#980 for moving those
        # relevant code to traitsui.
        foo = Foo(name="william", nickname="bill")
        view = View(
            Item("name", format_func=lambda s: s.upper(), style=style),
            Item("nickname", style=style),
        )
        tester = UITester()
        with tester.create_ui(foo, dict(view=view)) as ui:
            display_name = (
                tester.find_by_name(ui, "name").inspect(DisplayedText())
            )
            display_nickname = (
                tester.find_by_name(ui, "nickname").inspect(DisplayedText())
            )
            self.assertEqual(display_name, "WILLIAM")
            self.assertEqual(display_nickname, "bill")

    @unittest.skipUnless(
        Version(TRAITS_VERSION) >= Version("6.1.0"),
        "This test requires traits >= 6.1.0"
    )
    def test_format_func_used_simple(self):
        self.check_format_func_used(style='simple')

    @unittest.skipUnless(
        Version(TRAITS_VERSION) >= Version("6.1.0"),
        "This test requires traits >= 6.1.0"
    )
    def test_format_func_used_custom(self):
        self.check_format_func_used(style='custom')

    @unittest.skipUnless(
        Version(TRAITS_VERSION) >= Version("6.1.0"),
        "This test requires traits >= 6.1.0"
    )
    def test_format_func_used_readonly(self):
        self.check_format_func_used(style='readonly')



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.wx.file_editor import SimpleEditor
from traitsui.testing.tester._ui_tester_registry.wx._registry_helper import (
    register_editable_textbox_handlers,
)


def register(registry):
    """ Register interactions for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    register_editable_textbox_handlers(
        registry=registry,
        target_class=SimpleEditor,
        widget_getter=lambda wrapper: wrapper._target._file_name,
    )
